#include "control.h"
#include "gl_key.h"
#include "logfile.h"

using namespace GameLauncher;
using LogFile::gl_log;

#define _S(_A_, _B_) \
char * Emu_Menu_Controller:: _A_  = _B_

_S( CONFIG_SECTION,                     "Main");
_S( CONFIG_PREFER_INTEGER_SCALING,      "prefer_integer_scaling");
_S( CONFIG_SCREEN_SHOT_DELAY,           "screen_shot_delay");
_S( CONFIG_SCREEN_SHOT_BRIGHTNESS,      "screen_shot_brightness");
_S( CONFIG_SHOW_SCREEN_SHOT_ICON,       "show_screen_shot_icon");
_S( CONFIG_INITIAL_KEY_DELAY,           "initial_key_delay");
_S( CONFIG_KEY_REPEAT_DELAY,            "key_repeat_delay");
_S( CONFIG_INITIAL_JOYSTICK_DELAY,      "initial_joystick_delay");
_S( CONFIG_JOYSTICK_REPEAT_DELAY,       "joystick_repeat_delay");
_S( CONFIG_JOYSTICK_SELECT_BUTTON,      "joystick_select_button");
_S( CONFIG_UP_KEY,                      "up_key");
_S( CONFIG_DOWN_KEY,                    "down_key");
_S( CONFIG_PAGE_UP_KEY,                 "page_up_key");
_S( CONFIG_PAGE_DOWN_KEY,               "page_down_key");
_S( CONFIG_PLAY_GAME_KEY,               "play_game_key");
_S( CONFIG_NEXT_EMU_KEY,                "next_emulator_key");
_S( CONFIG_PREVIOUS_EMU_KEY,            "previous_emulator_key");
_S( CONFIG_QUIT_KEY,                    "quit_key");
_S( CONFIG_SCREEN_SHOT_KEY,             "screen_shot_key");
_S( CONFIG_MOUSE_INPUT_AXIS,            "mouse_input_axis");
_S( CONFIG_MOUSE_INPUT_THRESHOLD,       "mouse_input_threshold");
_S( CONFIG_USE_ANIMATIONS,              "use_animations");
_S( CONFIG_TURBO_SCROLL_THRESHOLD,      "turbo_scroll_threshold");
_S( CONFIG_LARGE_MENU_THRESHOLD,        "large_menu_threshold");
_S( CONFIG_EFFECT_VOLUME,               "sound_effect_volume");
_S( CONFIG_UNSELECTED_FG_COLOR,         "unselected_fg_color");
_S( CONFIG_UNSELECTED_OUTLINE_COLOR,    "unselected_outline_color");
_S( CONFIG_SELECTED_FG_COLOR,           "selected_fg_color");
_S( CONFIG_SELECTED_OUTLINE_COLOR,      "selected_outline_color");

#undef _S


Emu_Menu_Controller::Emu_Menu_Controller(
    Engine * engine, FONT * font, int x, int y, int width, int height,
    Menu_Type menu_type, Emulator_Manager * emulator_manager)
    : p_game_list()
{
    p_engine    = engine;
    p_font      = font;

    if (menu_type == HORIZONTAL)
    {
        p_menu = new HMenu_View(engine, font, x, y, width, height);
    }
    else if (menu_type == VERTICAL)
    {
        p_menu = new VLMenu_View(engine, font, x, y, width, height);
    }

    p_screen_shot       = new Screen_Shot(engine);
    p_screen_shot_icon  = 0;
    p_screen_shot_delay = 0;
    p_screen_shot_brightness    = 50;
    p_show_screen_shot_icon     = false;

    p_move_delay        = MOVE_DELAY;

    GEContainer *       root;
    root = p_engine->get_root_container();

    set_views_hidden(false);

    // Initialize samples to nothing
    p_next_emu          = 0;
    p_previous_emu      = 0;
    p_selection_changed = 0;
    p_game_selected     = 0;
    p_quit              = 0;

    p_null_listener     = new Null_Emu_Menu_Listener();
    p_listener          = p_null_listener;

    p_menu->set_listener(this);

    p_screen_shot->set_listener(this);

    p_emulator_manager  = emulator_manager;
    p_emulator_manager->set_listener(this);

    p_state     = INIT;
}

Emu_Menu_Controller::~Emu_Menu_Controller()
{
    GEContainer *       root;
    root = p_engine->get_root_container();

    root->remove_drawable(p_menu);
    root->remove_updatable(p_menu);

    root->remove_drawable(p_screen_shot);
    root->remove_updatable(p_screen_shot);

    delete p_menu;
    delete p_screen_shot;
}

void
Emu_Menu_Controller::play_sample(SAMPLE * sample)
{
    ::play_sample(sample, p_effect_volume, 127, 1000, FALSE);
}

string
Emu_Menu_Controller::get_current_screen_shot_file()
{
    Game *      game;
    Emulator *  emulator;
    string      screen_shot_file;

    emulator    = p_emulator_manager->get_current_emulator();
    screen_shot_file    = "";

    if (emulator != 0)
    {
        if (p_game_list.size() > 0)
        {
            game        = p_game_list[p_menu->get_selected_index()];
            screen_shot_file    = emulator->get_screen_shot_file_name(game);
        }
    }

    return screen_shot_file;
}

void
Emu_Menu_Controller::set_current_emulator_indecies()
{
    unsigned    selected;
    unsigned    first_shown;
    Emulator *  emulator;

    selected    = p_menu->get_selected_index();
    first_shown = p_menu->get_first_shown_index();

    emulator    = p_emulator_manager->get_current_emulator();
    if (emulator != 0)
    {
        emulator->set_selected_index(selected);
        emulator->set_first_shown_index(first_shown);
    }
}

void
Emu_Menu_Controller::start()
{
    p_menu->unhide_all(p_move_delay);
    p_state = MOVING_ON;
}

void
Emu_Menu_Controller::load_state(DataFile * data_file,
                                Config_File * config_file)
{
    bool        prefer_integer_scaling;

    p_next_emu          = data_file->change_game();
    p_previous_emu      = data_file->change_game();
    p_selection_changed = data_file->selection_moved();
    p_game_selected     = data_file->start_game();
    p_quit              = data_file->quit();
    p_screen_shot_icon  = data_file->screen_shot_icon();

    Config_Section *    section;
    int scancode;

    section     = config_file->get_section(CONFIG_SECTION);
    
    /* assign joystick events to menu actions */
    int joystick_select_button;
    joystick_select_button  =
       section->get_int(CONFIG_JOYSTICK_SELECT_BUTTON, BUTTON_A);
    p_menu->set_select_joy(joystick_select_button);
    p_menu->set_line_up_joy(JOY_UP);
    p_menu->set_line_down_joy(JOY_DOWN);
    p_menu->set_page_up_joy(JOY_LEFT);
    p_menu->set_page_down_joy(JOY_RIGHT);
    
    
    prefer_integer_scaling      =
        section->get_bool(CONFIG_PREFER_INTEGER_SCALING, true);
    p_screen_shot->set_prefer_integer_scaling(prefer_integer_scaling);

    p_screen_shot_delay = section->get_int(CONFIG_SCREEN_SHOT_DELAY, 0);
    p_screen_shot_brightness    =
        section->get_int(CONFIG_SCREEN_SHOT_BRIGHTNESS, 0);

    p_show_screen_shot_icon     =
        section->get_bool(CONFIG_SHOW_SCREEN_SHOT_ICON, false);

    scancode    = section->get_int(CONFIG_UP_KEY, GL_KEY_UP);
    scancode    = gl_key_to_allegro_key(scancode);
    p_menu->set_line_up_key(scancode);
   
    scancode    = section->get_int(CONFIG_DOWN_KEY, GL_KEY_DOWN);
    scancode    = gl_key_to_allegro_key(scancode);
    p_menu->set_line_down_key(scancode);
         
    scancode    = section->get_int(CONFIG_PAGE_UP_KEY, GL_KEY_LEFT);
    scancode    = gl_key_to_allegro_key(scancode);
    p_menu->set_page_up_key(scancode);

    scancode    = section->get_int(CONFIG_PAGE_DOWN_KEY, GL_KEY_RIGHT);
    scancode    = gl_key_to_allegro_key(scancode);
    p_menu->set_page_down_key(scancode);

    scancode    = section->get_int(CONFIG_PLAY_GAME_KEY, GL_KEY_1);
    scancode    = gl_key_to_allegro_key(scancode);
    p_menu->set_select_key(scancode);
   
    scancode    = section->get_int(CONFIG_NEXT_EMU_KEY, GL_KEY_ALT);
    scancode    = gl_key_to_allegro_key(scancode);
    p_menu->set_next_menu_key(scancode);

    scancode    = section->get_int(CONFIG_PREVIOUS_EMU_KEY, GL_KEY_LCONTROL);
    scancode    = gl_key_to_allegro_key(scancode);
    p_menu->set_previous_menu_key(scancode);

    scancode    = section->get_int(CONFIG_QUIT_KEY, GL_KEY_ESC);
    scancode    = gl_key_to_allegro_key(scancode);
    p_menu->set_quit_key(scancode);

    scancode    = section->get_int(CONFIG_SCREEN_SHOT_KEY, GL_KEY_SPACE);
    scancode    = gl_key_to_allegro_key(scancode);
    p_screen_shot->set_toggle_key(scancode);

    bool        use_animations;
    use_animations      = section->get_bool(CONFIG_USE_ANIMATIONS, true);
    if (use_animations)
    {
        p_move_delay    = MOVE_DELAY;
        p_menu->set_use_animations(true);
    }
    else
    {
        p_move_delay    = 0;
        p_menu->set_use_animations(0);
    }

    p_turbo_scroll_threshold  =
        section->get_int(CONFIG_TURBO_SCROLL_THRESHOLD, -1);
    p_menu->set_turbo_scroll_threshold(p_turbo_scroll_threshold);

    p_large_menu_threshold =
        section->get_int(CONFIG_LARGE_MENU_THRESHOLD, 0);

    p_mouse_input_axis  =
        section->get_int(CONFIG_MOUSE_INPUT_AXIS, 0);
    p_menu->set_mouse_input_axis(int_to_mouse_input_axis(p_mouse_input_axis));

    p_mouse_input_threshold =
        section->get_int(CONFIG_MOUSE_INPUT_THRESHOLD, 250);
    if (p_mouse_input_threshold < 1)
        p_mouse_input_threshold = 1;
    p_menu->set_mouse_input_threshold(p_mouse_input_threshold);
    
    p_effect_volume     = section->get_int(CONFIG_EFFECT_VOLUME, 1000);
    // Constrain to 0-1000;
    if (p_effect_volume < 0)
        p_effect_volume = 0;
    if (p_effect_volume > 1000)
        p_effect_volume = 1000;

    // Normalize to a value between 0-255
    p_effect_volume     = (p_effect_volume * 255) / 1000;

    int         initial_delay;
    int         repeat_delay;
    initial_delay       = section->get_int(CONFIG_INITIAL_KEY_DELAY, 250);
    repeat_delay        = section->get_int(CONFIG_KEY_REPEAT_DELAY, 125);
    p_menu->set_key_repeat_delay(initial_delay, repeat_delay);
  
    // Reuse initial_delay, repeat_delay for joystick 
    initial_delay       = section->get_int(CONFIG_INITIAL_JOYSTICK_DELAY, 250);
    repeat_delay        = section->get_int(CONFIG_JOYSTICK_REPEAT_DELAY, 125);
    p_menu->set_joystick_repeat_delay(initial_delay, repeat_delay);

    int unselected_fg_color             = makecol(255, 255, 255);
    int unselected_outline_color        = makecol(  0,   0,   0);
    int selected_fg_color               = makecol(  0,   0,   0);
    int selected_outline_color          = makecol(255, 255, 255);
    int r,g,b;
    vector<string> *    color;

    color       = section->get_string_list(CONFIG_UNSELECTED_FG_COLOR);
    if (color->size() == 3)
    {
        get_color(color, &r, &g, &b);
        unselected_fg_color = makecol(r, g, b);
    }
    delete color;

    color       = section->get_string_list(CONFIG_UNSELECTED_OUTLINE_COLOR);
    if (color->size() == 3)
    {
        get_color(color, &r, &g, &b);
        unselected_outline_color = makecol(r, g, b);
    }
    delete color;

    color       = section->get_string_list(CONFIG_SELECTED_FG_COLOR);
    if (color->size() == 3)
    {
        get_color(color, &r, &g, &b);
        selected_fg_color = makecol(r, g, b);
    }
    delete color;

    color       = section->get_string_list(CONFIG_SELECTED_OUTLINE_COLOR);
    if (color->size() == 3)
    {
        get_color(color, &r, &g, &b);
        selected_outline_color = makecol(r, g, b);
    }
    delete color;

    delete section;


    p_menu->set_selected_colors(selected_fg_color, selected_outline_color);
    p_menu->set_unselected_colors(unselected_fg_color,
                                  unselected_outline_color);

    // p_menu->set_selected_colors(makecol(0, 0, 0), makecol(192, 192, 64));
    // p_menu->set_unselected_colors(makecol(192, 192, 192), makecol(0, 0, 0));

    // Force a redraw as if the model has changed
    emulator_manager_changed();
}

void
Emu_Menu_Controller::get_color(vector<string> * color,
                               int * r, int * g, int * b)
{
    *r = *g = *b = 0;
    if (color->size() != 3)
        return;

    *r  = atoi((*color)[0].c_str());
    *g  = atoi((*color)[1].c_str());
    *b  = atoi((*color)[2].c_str());
}

Mouse_Input_Axis
Emu_Menu_Controller::int_to_mouse_input_axis(int int_axis)
{
    Mouse_Input_Axis    axis;
    
    switch (int_axis)
    {
        case 0:
            axis = NO_AXIS;
            break;

        case 1:
            axis = POSITIVE_X_AXIS;
            break;

        case 2:
            axis = NEGATIVE_X_AXIS;
            break;

        case 3:
            axis = POSITIVE_Y_AXIS;
            break;

        case 4:
            axis = NEGATIVE_Y_AXIS;
            break;

        default:
            axis = NO_AXIS;
            break;
    }

    return axis;
}

void
Emu_Menu_Controller::save_state(DataFile * data_file,
                                Config_File * config_file)
{
}

void
Emu_Menu_Controller::set_listener(Emu_Menu_Listener * listener)
{
    if (listener != 0)
        p_listener = listener;
    else
        p_listener = p_null_listener;
}

void
Emu_Menu_Controller::selection_changed(unsigned index)
{
    play_sample(p_selection_changed);
    p_screen_shot->cancel_set_bitmap();
}

void
Emu_Menu_Controller::item_selected(unsigned index)
{
    // Do nothing if there are no emulators
    if (p_emulator_manager->get_current_emulator() == 0)
        return;

    p_state = GAME_SELECTED;
    play_sample(p_game_selected);
    p_menu->center_selected_hide_others(p_move_delay);
    p_screen_shot->cancel_set_bitmap();
    set_current_emulator_indecies();
}

void
Emu_Menu_Controller::item_visible(unsigned index)
{
    Game *      game;
    Emulator *  emulator;
    string      screen_shot_file;

    emulator    = p_emulator_manager->get_current_emulator();
    game        = p_game_list[index];
    screen_shot_file    = "";

    if (emulator != 0)
        screen_shot_file    = emulator->get_screen_shot_file_name(game);

    if (screen_shot_file != "")
    {
        Scrollable_Item item;
        item.text       = p_game_list[index]->get_full_name();
        if (p_show_screen_shot_icon)
            item.icon   = p_screen_shot_icon;
        else
            item.icon   = 0;

        p_menu->set_item_at(item, index);
    }
}

void
Emu_Menu_Controller::quit_pressed()
{
    p_state = QUIT_SELECTED;
    play_sample(p_quit);
    p_menu->hide_split(p_move_delay);
    p_screen_shot->cancel_set_bitmap();
    set_current_emulator_indecies();
}

void
Emu_Menu_Controller::next_menu_pressed()
{
    // Do nothing if there are no emulators
    if (p_emulator_manager->get_current_emulator() == 0)
        return;

    play_sample(p_next_emu);
    p_state = NEXT_EMU;
    p_menu->hide_right(p_move_delay);
    p_screen_shot->cancel_set_bitmap();
    set_current_emulator_indecies();
}

void
Emu_Menu_Controller::previous_menu_pressed()
{
    // Do nothing if there are no emulators
    if (p_emulator_manager->get_current_emulator() == 0)
        return;

    play_sample(p_previous_emu);
    p_state = PREVIOUS_EMU;
    p_menu->hide_left(p_move_delay);
    p_screen_shot->cancel_set_bitmap();
    set_current_emulator_indecies();
}

void
Emu_Menu_Controller::stopped_moving()
{
    if (p_state == MOVING_ON)
    {
        string      screen_shot_file;

        p_state = SELECTING;
        screen_shot_file    = get_current_screen_shot_file();
        p_screen_shot->set_bitmap(screen_shot_file, p_screen_shot_brightness);
    }
    else if (p_state == NEXT_EMU)
    {
        p_state = MOVING_ON;
        p_menu->hide_left(0);

        p_emulator_manager->increment_emulator_index();

        p_menu->unhide_all(p_move_delay);
    }
    else if (p_state == PREVIOUS_EMU)
    {
        p_state = MOVING_ON;
        p_menu->hide_right(0);
        
        p_emulator_manager->decrement_emulator_index();

        p_menu->unhide_all(p_move_delay);
    }
    else if (p_state == QUIT_SELECTED)
    {
        p_state = DONE;
        p_listener->quit_selected();
    }
    else if (p_state == GAME_SELECTED)
    {
        Emulator *      emulator;
        Game *          game;

        p_state = DONE;
        emulator        = p_emulator_manager->get_current_emulator();
        if (emulator != 0)
        {
            game = p_game_list[emulator->get_selected_index()];
            p_listener->game_selected(emulator, game);
        }
    }
    else if (p_state == SELECTING)
    {
        string      screen_shot_file;

        screen_shot_file    = get_current_screen_shot_file();
        p_screen_shot->set_bitmap(screen_shot_file, p_screen_shot_brightness,
                                  p_screen_shot_delay);
    }
}

void
Emu_Menu_Controller::toggle_screen_shot()
{
    string      screen_shot_file;

    screen_shot_file    = get_current_screen_shot_file();

    if (p_state == SELECTING)
    {
        if (screen_shot_file == "")
            return;

        GEContainer *   root;
        root    = p_engine->get_root_container();

        root->remove_drawable(p_menu);
        root->remove_updatable(p_menu);
        p_screen_shot->set_bitmap(screen_shot_file, 100);
        p_state = DISPLAYING_SCREEN_SHOT;
    }
    else if (p_state == DISPLAYING_SCREEN_SHOT)
    {
        GEContainer *   root;
        root    = p_engine->get_root_container();

        root->add_drawable(p_menu, DRAW_PRIORITY);
        root->add_updatable(p_menu, UPDATE_PRIORITY);
        p_screen_shot->set_bitmap(screen_shot_file, p_screen_shot_brightness);
        p_state = SELECTING;
    }
}

void
Emu_Menu_Controller::emulator_manager_changed()
{
    Emulator *  emulator;
    int number_of_games;
    string      title;
    string      sub_title;
    unsigned    selected_index;
    unsigned    first_shown_index;

    number_of_games     = 0;
    title               = "No Emulators Loaded";
    sub_title           = "(Please Reconfigure)";
    selected_index      = 0;
    first_shown_index   = 0;
    p_game_list.resize(0);

    emulator    = p_emulator_manager->get_current_emulator();
    if (emulator != 0)
    {
        emulator->get_game_list(p_game_list);
        number_of_games = p_game_list.size();
        title           = emulator->name();
        sub_title       = emulator->description();
        selected_index  = emulator->get_selected_index();
        first_shown_index       = emulator->get_first_shown_index();
    }

    Scrollable_Item_List        item_list(number_of_games);

    Game_List::iterator                 i;
    Scrollable_Item_List::iterator      j;

    for (i = p_game_list.begin(), j = item_list.begin();
         i != p_game_list.end();
         i++, j++)
    {
        (*j).text       = (*i)->get_full_name();
        (*j).icon       = 0;
    }
    p_menu->set_titles(title, sub_title);
    p_menu->set_all_items(item_list, selected_index, first_shown_index);

    // Enable turbo scrolling if the number of games in this menu has
    // passed the threshold of a "large menu".
    if (number_of_games > p_large_menu_threshold)
        p_menu->set_turbo_scroll_threshold(p_turbo_scroll_threshold);
    else
        p_menu->set_turbo_scroll_threshold(-1);
        
    p_engine->catchup();
}

void
Emu_Menu_Controller::poll_emulator_manager()
{
    p_engine->poll_background_task();
}

void
Emu_Menu_Controller::set_views_hidden(bool hidden)
{
    GEContainer *   root;
    root    = p_engine->get_root_container();

    if (hidden)
    {
        root->remove_drawable(p_menu);
        root->remove_updatable(p_menu);

        root->remove_drawable(p_screen_shot);
        root->remove_updatable(p_screen_shot);
    }
    else
    {
        if (p_state != DISPLAYING_SCREEN_SHOT)
        {
            root->add_drawable(p_menu, DRAW_PRIORITY);
            root->add_updatable(p_menu, UPDATE_PRIORITY);
        }

        root->add_drawable(p_screen_shot, 0);
        root->add_updatable(p_screen_shot, UPDATE_PRIORITY);
    }
}


/************************************************************************
 *
 * Null_Emu_Menu_Listener
 *
 ************************************************************************/

void
Null_Emu_Menu_Listener::quit_selected()
{
}

void
Null_Emu_Menu_Listener::game_selected(Emulator * emulator, Game * game)
{
}
