#include "allegro.h"
#include "fileutil.h"
#include "logfile.h"

#if defined(TARGET_MINGW)
#include "winalleg.h"
#endif

#if defined(TARGET_UNIX) || defined(TARGET_DJGPP)
#include <sys/stat.h>
#include <sys/types.h>
#endif

using LogFile::gl_log;

inline static bool is_separator(char ch)
{
    return ((ch == '/') || (ch == '\\'));
}

string  get_extension(const string & path)
{
    string::size_type   i;
    string      extension;
    bool        found;
    bool        done;

    // Start at the end of the string and work our way in
    i           = path.length() - 1;
    found       = false;
    done        = false;
    while (! done)
    {
        char    ch;

        ch = path[i];
        if (ch == '.')
        {
            found = true;
            break;
        }

        i--;

        // Not found, so just break
        if (is_separator(ch) || (i == 0))
            done = true;
    }
    
    if (found)
        // Add one to skip the "."
        return string(path, i+1);
    else
        return string("");
}

string  remove_extension(const string & path)
{
    string::size_type   i;
    string      extension;
    bool        found;
    bool        done;

    i           = path.length() - 1;
    found       = false;
    done        = false;

    while (! done)
    {
        char    ch;

        ch = path[i];
        if (ch == '.')
        {
            found = true;
            break;
        }

        // Not found, so just break
        if ((is_separator(ch)) || (i == 0))
            done = true;

        i--;
    }
    
    if (found)
        return string(path, 0, i);
    else
        return path;
}

void    split_extension(const string & path, string & base_name,
                        string & extension)
{
    string::size_type   i;
    bool        found;
    bool        done;

    i           = path.length() - 1;
    found       = false;
    done        = false;

    while (! done)
    {
        char    ch;

        ch = path[i];
        if (ch == '.')
        {
            found = true;
            break;
        }

        // Not found, so just break
        if ((is_separator(ch)) || (i == 0))
            done = true;

        i--;
    }
    
    if (found)
    {
        base_name       = string(path, 0, i);
        extension       = string(path, i+1);
    }
    else
    {
        base_name       = path;
        extension       = "";
    }
}

void    add_backslash(string & path)
{
    if (! is_separator(path[path.length()-1]))
        path += "\\";
    return;
}

void    fix_filename_slashes(string & path)
{
    string::size_type   i;

    for (i = 0; i < path.length(); i++)
    {
#if defined(TARGET_MINGW) || defined(TARGET_DJGPP)
        if (path[i] == '/')
            path[i] = '\\';
#else
        if (path[i] == '\\')
            path[i] = '/';
#endif
    }
    return;
}

bool    is_absolute_path(const string & path)
{
    // Absolute path begins with driver letter (x:...), '\' or '/'
    if (path[1] == ':')
        return true;
    else if (path[0] == '/')
        return true;
    else if (path[0] == '\\')
        return true;

    return false;
}

void    quote_long_file_name(string & file_name)
{
    unsigned    i;
    bool        add_quotes;

    add_quotes  = false;

    for (i = 0; i < file_name.length(); i++)
    {
        char    ch;

        ch = file_name[i];

        if (ch == ' ')
        {
            add_quotes  = true;
            break;
        }
    }

    if (add_quotes)
        file_name       = "\"" + file_name + "\"";
    return;
}

void fix_file_name_case(string & file_name)
{
    string::iterator  i;

    // Change to upper case, if needed.
    if (! ALLEGRO_LFN)
    {
        i   = file_name.begin();

        while (i != file_name.end())
        {
            *i = toupper(*i);
            i++;
        }
    }
}

void    set_config_file(const string & config_file)
{
    set_config_file(config_file.c_str());
}

string  get_config_string(const string & section, const string & name,
                          const string & default_value)
{
    return string(get_config_string(section.c_str(), name.c_str(),
                                    default_value.c_str()));
}

void    set_config_string(const string & section, const string & name,
                          const string & value)
{
    set_config_string(section.c_str(), name.c_str(), value.c_str());
}

bool
get_config_bool(const string & section, const string & name,
                bool default_value)
{
    string      string_bool;
    if (default_value == true)
        string_bool = get_config_string(section, name, "yes");
    else
        string_bool = get_config_string(section, name, "no");

    if (string_bool == "yes")
        return true;
    else
        return false;
}

void
set_config_bool(const string & section, const string & field, bool value)
{
    string      string_bool;

    if (value == true)
        string_bool     = "yes";
    else
        string_bool     = "no";

    set_config_string(section, field, string_bool);
}

int
get_config_int(const string & section, const string & field, int default_value)
{
    return get_config_int(section.c_str(), field.c_str(), default_value);
}

void
set_config_int(const string & section, const string & field, int value)
{
    set_config_int(section.c_str(), field.c_str(), value);
}

char **
get_config_argv(const string & section, const string & field, int * argc)
{
    return get_config_argv(section.c_str(), field.c_str(), argc);
}

bool
file_exists(const string & file)
{
    return exists(file.c_str());
}

bool
directory_exists(const string & directory)
{
    int aret;
    bool exists = file_exists(directory.c_str(), FA_DIREC, &aret);
    if (exists)
    {
        if ((aret & FA_DIREC) != 0)
            return true;
        else
            return false;
    }
    return false;
}

bool create_directory(const string & directory)
{
#if defined(TARGET_MINGW)
    return CreateDirectory((char *) directory.c_str(), 0);
#elif defined(TARGET_UNIX) || defined(TARGET_DJGPP)
    if (mkdir(directory.c_str(), (S_IRWXU | S_IRWXG | S_IRWXO)) < 0)
        return false;
    else
        return true;
#else
#error Invalid Target
#endif
}

long
file_size(const string & file)
{
    return file_size(file.c_str());
}

Config_File::Config_File(const string & file_name)
{
    p_file_name = file_name;
}

Config_Section *
Config_File::get_section(const string & section)
{
    push_config_state();
    set_config_file(p_file_name);

    Config_Section *    new_section;
    new_section = new Config_Section_Impl(section);

    return new_section;
}

Config_Section::~Config_Section()
{
}

Config_Section_Impl::Config_Section_Impl(const string & section)
{
    p_section   = section;
}

Config_Section_Impl::~Config_Section_Impl()
{
    pop_config_state();
}

string
Config_Section_Impl::get_string(const string & field,
                                const string & default_value)
{
    return get_config_string(p_section, field, default_value);
}

void
Config_Section_Impl::set_string(const string & field,
                                const string & value)
{
    set_config_string(p_section, field, value);
}

string
Config_Section_Impl::get_file_name(const string & field,
                                   const string & default_value)
{
    string file_name;

    file_name = get_string(field, default_value);
    fix_file_name_case(file_name);
    return file_name;
}

bool
Config_Section_Impl::get_bool(const string & field, bool default_value)
{
    return get_config_bool(p_section, field, default_value);
}

void
Config_Section_Impl::set_bool(const string & field, bool value)
{
    set_config_bool(p_section, field, value);
}
    
int
Config_Section_Impl::get_int(const string & field, int default_value)
{
    return get_config_int(p_section, field, default_value);
}

void
Config_Section_Impl::set_int(const string & field, int value)
{
    set_config_int(p_section, field, value);
}

vector<string> *
Config_Section_Impl::get_string_list(const string & field)
{
    vector<string> *    string_list;
    char **     argv;
    int         argc;
    int         i;

    string_list = new vector<string>;

    argv        = get_config_argv(p_section, field, &argc);

    string_list->resize(argc);
    for (i = 0; i < argc; i++)
    {
        (*string_list)[i] = string(argv[i]);
    }

    return string_list;
}

vector<string> *
Config_Section_Impl::get_file_name_list(const string & field)
{
    vector<string> *    file_name_list;
    vector<string>::iterator i;

    file_name_list = get_string_list(field);
    i = file_name_list->begin();
    while (i != file_name_list->end())
    {
        fix_file_name_case(*i);
        i++;
    }

    return file_name_list;
}
