#include "enginei.h"
#include "logfile.h"

using namespace GameEngine;
using namespace std;
using LogFile::gl_log;


Keypress_Watcher_Impl::Keypress_Watcher_Impl(Keypress_Manager *
                                             keypress_manager,
                                             Timer * timer)
{
    p_type      = CONTINUOUS;
    p_scancode  = 0;
    p_is_pressed        = false;
    p_was_pressed       = false;
    p_initial_delay     = 0;
    p_repeat_delay      = 0;
    p_timer             = timer;
    p_keypress_manager  = keypress_manager;
}

Keypress_Watcher_Impl::~Keypress_Watcher_Impl()
{
    delete p_timer;
    p_timer = 0;
    p_keypress_manager->delete_keypress_watcher(this);
}

void
Keypress_Watcher_Impl::set_scancode(int scancode)
{
    p_scancode  = scancode;
}

bool
Keypress_Watcher_Impl::is_pressed()
{
    return p_is_pressed;
}

bool
Keypress_Watcher_Impl::is_physically_pressed()
{
    return key[p_scancode];
}

bool
Keypress_Watcher_Impl::is_artificial()
{
    return p_is_artificial;
}

void
Keypress_Watcher_Impl::watch()
{
    // Key is not pressed and was not pressed
    if ((! key[p_scancode]) && (! p_was_pressed))
        return;

    // Key was just released, so reset state
    if (! key[p_scancode])
    {
        reset_state();
        return;
    }

    if (p_type == CONTINUOUS)
    {
        p_is_pressed        = true;
        if (p_was_pressed)
        {
            p_is_artificial = true;
        }
    }
    else if (p_type == ONE_TIME)
    {
        if (! p_was_pressed)
            p_is_pressed    = true;
        else        
            p_is_pressed    = false;
    }
    else if (p_type == REPEAT)
    {
        if (! p_was_pressed)
        {
            // Initial keypress
            p_is_pressed    = true;
            p_is_artificial = false;
            p_timer->set(p_initial_delay);
        }
        else if (p_timer->expired())
        {
            // Repeat
            p_is_pressed    = true;
            p_is_artificial = true;
            p_timer->set(p_repeat_delay);
                
        }
        else
        {
            // In between repeats
            p_is_pressed    = false;
            p_is_artificial = false;
        }
    }

    p_was_pressed = true;
    return;
}

void
Keypress_Watcher_Impl::set_continuous()
{
    reset_state();
    p_type      = CONTINUOUS;
}

void
Keypress_Watcher_Impl::set_one_time()
{
    reset_state();
    p_type      = ONE_TIME;
}

void
Keypress_Watcher_Impl::set_repeat(int initial_delay, int repeat_delay)
{
    reset_state();
    p_initial_delay     = initial_delay;
    p_repeat_delay      = repeat_delay;
    p_type              = REPEAT;
}

void
Keypress_Watcher_Impl::reset_state()
{
    p_is_pressed    = false;
    p_is_artificial = false;
    p_was_pressed   = false;
    if (p_type == REPEAT)
        p_timer->disable();
}


Keypress_Manager::Keypress_Manager(Engine * engine)
{
    p_engine = engine;
}

Keypress_Manager::~Keypress_Manager()
{
}

Keypress_Watcher *
Keypress_Manager::create_keypress_watcher(int scancode)
{
    Keypress_Watcher_Impl *     keypress_watcher;
    Timer *                     timer;

    timer = p_engine->get_timer_factory()->create_timer();
    keypress_watcher = new Keypress_Watcher_Impl(this, timer);
    keypress_watcher->set_scancode(scancode);
    p_keypress_watcher_list.push_back(keypress_watcher);

    return keypress_watcher;
}

void
Keypress_Manager::delete_keypress_watcher(Keypress_Watcher * keypress_watcher)
{
    Keypress_Watcher_List::iterator     i;

    for (i =  p_keypress_watcher_list.begin();
         i != p_keypress_watcher_list.end();
         i++)
    {
        if (*i == keypress_watcher)
        {
            p_keypress_watcher_list.erase(i);
            break;
        }
    }

    return;
}

bool
Keypress_Manager::update()
{
    Keypress_Watcher_List::iterator     i;

    for (i =  p_keypress_watcher_list.begin();
         i != p_keypress_watcher_list.end();
         i++)
    {
        (*i)->watch();
    }

    // No visual updates
    return false;
}
