#include "outline.h"
#include "gui/uimgr.h"
#include "gui/label.h"
#include "gui/icon.h"
#include "view.h"
#include "allegro.h"

using namespace GUI;
using namespace Outline;


/************************************************************************
 *
 * Label
 *
 ************************************************************************/

Label::Label(int maximum_width, int maximum_height)
    : p_location(), p_outline_color()
{
    p_max_width = maximum_width;
    p_max_height = maximum_height;
    p_text = "";
    p_icon = 0;

    p_foreground_color = makecol(255, 255, 255);
    p_background_color = makecol(255, 0, 255);
    set_outline_color(makecol(0, 0, 0), 0);
    p_updated = false;

    p_ui = UI_Manager::get_label_ui();
    p_ui->install_ui(this);

    p_font = font;
}

Label::~Label()
{
    delete p_ui;
}

int
Label::get_maximum_width() const
{
    return p_max_width;
}

int
Label::get_maximum_height() const
{
    return p_max_height;
}

void
Label::set_location(int x, int y, int milliseconds = 0)
{
    p_location.move_to(x, y, milliseconds);
    p_updated = true;
}

bool
Label::is_stopped() const
{
    return p_location.is_stopped();
}

int
Label::get_x() const
{
    return p_location.get_x();
}

int
Label::get_y() const
{
    return p_location.get_y();
}

void
Label::set_size(int width, int height)
{
    p_width = width;
    p_height = height;
}

int
Label::get_width() const
{
    return p_width;
}

int
Label::get_height() const
{
    return p_height;
}

void
Label::set_contents(const string & text, const Icon * icon)
{
    p_text = text;
    p_icon = icon;
    p_updated = true;
}

string
Label::get_text() const
{
    return p_text;
}

const Icon *
Label::get_icon() const
{
    return p_icon;
}

void
Label::set_foreground_color(int foreground_color)
{
    p_foreground_color = foreground_color;
    p_updated = true;
}

int
Label::get_foreground_color()
{
    return p_foreground_color;
}

void
Label::set_background_color(int background_color)
{
    p_background_color = background_color;
    p_updated =true;
}

int
Label::get_background_color()
{
    return p_background_color;
}

void
Label::set_outline_color(int outline_color, int pulse_period)
{
    p_outline_color.set_color(outline_color, pulse_period);
    p_updated = true;
}

int
Label::get_outline_color()
{
    return p_outline_color.get_color();
}

FONT *
Label::get_font() const
{
    return UI_Manager::get_font();
}

bool
Label::update()
{
    if (p_outline_color.update())
    {
        p_updated = true;
    }

    // It doesn't matter if the location changed or not, so just
    // ignore the return code.  The UI doesn't need to be updated just
    // because the location changed.
    (void) p_location.update();

    if (p_updated)
    {
        p_ui->label_updated();
        p_updated = false;
    }

    return true;
}


void
Label::draw(BITMAP * frame)
{
    p_ui->draw(frame);
}


/************************************************************************
 *
 * Label_UI
 *
 ************************************************************************/

Label_UI::~Label_UI()
{
}


/************************************************************************
 *
 * Abstract_Label_UI
 *
 ************************************************************************/

Abstract_Label_UI::Abstract_Label_UI()
{
    p_main_bitmap = 0;
    p_label_bitmap = 0;
}

Abstract_Label_UI::~Abstract_Label_UI()
{
    destroy_bitmaps();
}

void
Abstract_Label_UI::destroy_bitmaps()
{
    if (p_main_bitmap == 0)
    {
        destroy_bitmap(p_main_bitmap);
        p_main_bitmap = 0;
    }

    if (p_label_bitmap == 0)
    {
        destroy_bitmap(p_label_bitmap);
        p_label_bitmap = 0;
    }
}

int
Abstract_Label_UI::truncate_text(string & text, FONT * font,
                                 int maximum_width)
{
    int text_width;
    
    text_width = outline_text_length(font, text);
    while (text_width > maximum_width)
    {
        // Chop one character off the end
        text.erase(text.length() - 1, text.length());
        text_width = outline_text_length(font, text);
    }

    return text_width;
}

void
Abstract_Label_UI::draw(BITMAP * bitmap)
{
    draw_sprite(bitmap, p_label_bitmap, p_label->get_x(), p_label->get_y());
}

BITMAP *
Abstract_Label_UI::draw_horizontal(Destination_Bitmap destination_bitmap,
                                   int maximum_width, int height)
{
    int max_text_width;
    int actual_text_width;
    int icon_width;
    int width;
    string text;
    const Icon * icon;
    FONT * label_font;
    BITMAP * destination;

    if (p_label_bitmap != 0)
    {
        destroy_bitmap(p_label_bitmap);
        p_label_bitmap = 0;
    }
    
    label_font = p_label->get_font();
    icon = p_label->get_icon();
    text = p_label->get_text();

    max_text_width = maximum_width;
    if (icon != 0)
    {
        icon_width = icon->get_icon_width();
        max_text_width -= icon_width;
    }
    else
    {
        icon_width = 0;
    }

    actual_text_width = truncate_text(text, label_font, max_text_width);
    width = actual_text_width + icon_width;
    
    if (destination_bitmap == SUB_BITMAP)
    {
        destination = create_sub_bitmap(p_main_bitmap, 0, 0, width, height);
    }
    else // destination_bitmap == TEMPORARY_BITMAP
    {
        destination = create_bitmap(width, height);
    }
    
    clear_to_color(destination, p_label->get_background_color());
    outline_textout(destination, label_font, text,
                    p_label->get_foreground_color(),
                    p_label->get_outline_color());

    if (icon != 0)
    {
        int icon_y;

        // Try to center icon around horizontal axis
        icon_y = (destination->h / 2) - (icon->get_icon_height() / 2);
        if (icon_y < 0)
        {
            // If icon height is large than bitmap height, just put it
            // at the top of the bitmap and the clipping will truncate
            // the bottom.
            icon_y = 0;
        }
        icon->paint_icon(destination, actual_text_width, icon_y);
    }

    return destination;
}
