#include "engine.h"
#include "gui/uimgr.h"
#include "gui/util.h"

using GameEngine::Engine;
using namespace GUI;


/************************************************************************
 *
 * Movable_Point
 *
 ************************************************************************/

Movable_Point::Movable_Point()
{
    p_x = 0;
    p_y = 0;
    p_velocity_x = 0;
    p_velocity_y = 0;
    p_destination_x = 0;
    p_destination_y =0;
    p_state = STOPPED;
}

void
Movable_Point::move_to(int x, int y, int milliseconds)
{
    fix       num_game_ticks;
    Engine * engine;

    if (milliseconds == 0)
    {
        p_x = x;
        p_y = y;
        p_state = STOPPED;
        return;
    }

    engine = UI_Manager::get_engine();
    num_game_ticks = (milliseconds * engine->updates_per_second()) / 1000;

    p_destination_x = x;
    p_destination_y = y;
    p_velocity_x = (p_destination_x - p_x) / num_game_ticks;
    p_velocity_y = (p_destination_y - p_y) / num_game_ticks;

    // If destination is so close such that the velocity is zero (with
    // rounding), stay stopped
    if ((p_velocity_x == 0) && (p_velocity_y == 0))
    {
        p_state = STOPPED;
    }
    else
    {
        p_state = MOVING;
    }
}

int
Movable_Point::get_x() const
{
    return p_x;
}

int
Movable_Point::get_y() const
{
    return p_y;
}

bool
Movable_Point::is_stopped() const
{
    return (p_state == STOPPED);
}

bool
Movable_Point::update()
{
    bool    moving_right;
    bool    moving_down;

    if (p_state == STOPPED)
    {
        return false;
    }

    // Will we reach or pass our final destination?
    moving_right    = (p_velocity_x > 0);
    moving_down     = (p_velocity_y > 0);

    if ((moving_right && ((p_x + p_velocity_x) >= p_destination_x)) ||
        ((! moving_right) && ((p_x + p_velocity_x) <= p_destination_x)))
    {
        p_x = p_destination_x;
    }
    else
    {
        p_x += p_velocity_x;
    }

    if ((moving_down && ((p_y + p_velocity_y) >= p_destination_y)) ||
        ((! moving_down) && (p_y + p_velocity_y) <= p_destination_y))
    {
        p_y = p_destination_y;
    }
    else
    {
        p_y += p_velocity_y;
    }

    if ((p_x == p_destination_x) && (p_y == p_destination_y))
    {
        p_state = STOPPED;
    }
    
    return true;
}


/************************************************************************
 *
 * Pulsating_Color
 *
 ************************************************************************/

Pulsating_Color::Pulsating_Color()
{
    set_color(makecol(255, 255, 255), 0);
}

void
Pulsating_Color::set_color(int color, int pulse_period)
{
    fix num_ticks;
    Engine * engine;
    
    p_base_r = getr(color);
    p_base_g = getg(color);
    p_base_b = getb(color);

    p_current_r = p_base_r;
    p_current_g = p_base_g;
    p_current_b = p_base_b;

    // The fixed type does some goofy rounding if period is zero and
    // the delta's don't end up being zero, so this forces them to be
    // zero.
    if (pulse_period == 0)
    {
        p_delta_r = 0;
        p_delta_g = 0;
        p_delta_b = 0;
    }
    else
    {
        engine = UI_Manager::get_engine();
        num_ticks = ((pulse_period/2) * engine->updates_per_second()) / 1000;
        p_delta_r = p_base_r / num_ticks;
        p_delta_g = p_base_g / num_ticks;
        p_delta_b = p_base_b / num_ticks;
        p_direction = -1;
    }
}

int
Pulsating_Color::get_color() const
{
    return makecol(int(p_current_r), int(p_current_g), int(p_current_b));
}

bool
Pulsating_Color::update()
{
    fix last_r;
    fix last_g;
    fix last_b;
    bool updated;
    bool switch_direction;

    last_r = p_current_r;
    last_g = p_current_g;
    last_b = p_current_b;
    updated = false;
    switch_direction = false;
    

    if (p_delta_r > 0)
    {
        p_current_r += p_delta_r * p_direction;
        updated = true;
        if (p_current_r > p_base_r)
        {
            p_current_r = p_base_r;
            switch_direction = true;
        }

        if (p_current_r < 0)
        {
            p_current_r = 0;
            switch_direction = true;
        }
    }

    if (p_delta_g > 0)
    {
        p_current_g += p_delta_g * p_direction;
        updated = true;
        if (p_current_g >= p_base_g)
        {
            p_current_g = p_base_g;
            switch_direction = true;
        }
        
        if (p_current_g < 0)
        {
            p_current_g = 0;
            switch_direction = true;
        }
    }

    
    if (p_delta_b > 0)
    {
        p_current_b += p_delta_b * p_direction;
        updated = true;
        if (p_current_b >= p_base_b)
        {
            p_current_b = p_base_b;
            switch_direction = true;
        }
        
        if (p_current_b < 0)
        {
            p_current_b = 0;
            switch_direction = true;
        }
    }

    if (switch_direction)
    {
        p_direction *= -1;
    }
    
    return updated;
}


/************************************************************************
 *
 * Rectangle
 *
 ************************************************************************/

Rectangle::Rectangle()
{
    set_bounds(0, 0, 1, 1);
}

Rectangle::Rectangle(int x, int y, int width, int height)
{
    set_bounds(x, y, width, height);
}

void
Rectangle::set_bounds(int x, int y, int width, int height)
{
    p_x = x;
    p_y = y;
    p_width = width;
    p_height = height;

    if (width <= 0)
        width = 1;

    if (height <= 0)
        height = 1;
}

int
Rectangle::get_x() const
{
    return p_x;
}

int
Rectangle::get_y() const
{
    return p_y;
}

int
Rectangle::get_lower_right_x() const
{
    return p_x + p_width - 1;
}

int
Rectangle::get_lower_right_y() const
{
    return p_y + p_height - 1;
}

int
Rectangle::get_width() const
{
    return p_width;
}

int
Rectangle::get_height() const
{
    return p_height;
}
