#ifndef _CONTROL_H_
#define _CONTROL_H_

#include <vector>
#include <list>
#include "glaunch.h"
#include "engine.h"
#include "view.h"
#include "fileutil.h"
#include "sshot.h"
#include "logfile.h"
#include "exit.h"

namespace GameLauncher
{
    using LogFile::Log_Rotate_Listener;
    using LogFile::Log_Rotate_Event;

    class Launcher_Controller;
    class Emu_Menu_Controller;
    class Emu_Menu_Listener;
    class Null_Emu_Menu_Listener;
    class Background_Music_Controller;
    class CD_Background;
    class MP3_Background;

    enum Menu_Type { HORIZONTAL, VERTICAL };

    class Emu_Menu_Controller : public Menu_Listener,
                                public Screen_Shot_Listener,
                                public Emulator_Manager_Listener
    {
      public:
        Emu_Menu_Controller(Engine * engine, FONT * font, int x, int y,
                            int width, int height, Menu_Type menu_type,
                            Emulator_Manager * emulator_manager);
        virtual ~Emu_Menu_Controller();

        void            load_state(DataFile * data_file,
                                   Config_File * config_file);
        void            save_state(DataFile * data_file,
                                   Config_File * config_file);

        void            set_listener(Emu_Menu_Listener * listener);
        void            start();

        // Menu_Listener interface
        void            selection_changed(unsigned index);
        void            item_selected(unsigned index);
        void            item_visible(unsigned index);
        void            quit_pressed();
        void            next_menu_pressed();
        void            previous_menu_pressed();
        void            stopped_moving();

        // Screen_Shot_Listener interface
        virtual void toggle_screen_shot();

        // Emulator_Manager_Listener interface
        virtual void emulator_manager_changed();
        virtual void poll_emulator_manager();

        /**
         * Sets whether or not all view objects are hidden.
         *
         * @param hidden True if views should be hidden.
         */
        void set_views_hidden(bool hidden);
        
      private:
        enum State {INIT,
                    MOVING_ON,
                    SELECTING,
                    NEXT_EMU,
                    PREVIOUS_EMU,
                    QUIT_SELECTED,
                    GAME_SELECTED,
                    DISPLAYING_SCREEN_SHOT,
                    DONE};
        static const int        DRAW_PRIORITY           = 400;
        static const int        UPDATE_PRIORITY         = 400;
        static const int        MOVE_DELAY              = 500;

        void            play_sample(SAMPLE * sample);
        string          get_current_screen_shot_file();
        void            set_current_emulator_indecies();
        void            get_color(vector<string> * color,
                                  int * r, int * g, int * b);

        /**
         * Converts an integer to a Mouse_Input_Axis.  If the integer
         * does not map, NO_AXIS is returned.
         *
         * @param int_axis Integer representation of mouse input axis.
         * @return The corresponding mouse input axis.
         */
        Mouse_Input_Axis int_to_mouse_input_axis(int int_axis);

        Engine *        p_engine;
        FONT *          p_font;

        Emulator_Manager *      p_emulator_manager;
        Game_List       p_game_list;

        State           p_state;
        Emu_Menu_Listener *     p_listener;
        Emu_Menu_Listener *     p_null_listener;

        SAMPLE *        p_next_emu;
        SAMPLE *        p_previous_emu;
        SAMPLE *        p_selection_changed;
        SAMPLE *        p_game_selected;
        SAMPLE *        p_quit;
        int             p_effect_volume;

        BITMAP *        p_screen_shot_icon;
        bool            p_show_screen_shot_icon;
        int             p_screen_shot_delay;
        int             p_screen_shot_brightness;
        int             p_move_delay;
        int             p_turbo_scroll_threshold;
        int             p_large_menu_threshold;
        int             p_mouse_input_axis;
        int             p_mouse_input_threshold;

        Menu_View *     p_menu;
        Screen_Shot *   p_screen_shot;

        static char *   CONFIG_SECTION;
        static char *   CONFIG_PREFER_INTEGER_SCALING;
        static char *   CONFIG_SCREEN_SHOT_DELAY;
        static char *   CONFIG_SCREEN_SHOT_BRIGHTNESS;
        static char *   CONFIG_SHOW_SCREEN_SHOT_ICON;
        static char *   CONFIG_INITIAL_KEY_DELAY;
        static char *   CONFIG_KEY_REPEAT_DELAY;
        static char *   CONFIG_JOYSTICK_SELECT_BUTTON;
        static char *   CONFIG_INITIAL_JOYSTICK_DELAY;
        static char *   CONFIG_JOYSTICK_REPEAT_DELAY;
        static char *   CONFIG_UP_KEY;
        static char *   CONFIG_DOWN_KEY;
        static char *   CONFIG_PAGE_UP_KEY;
        static char *   CONFIG_PAGE_DOWN_KEY;
        static char *   CONFIG_PLAY_GAME_KEY;
        static char *   CONFIG_NEXT_EMU_KEY;
        static char *   CONFIG_PREVIOUS_EMU_KEY;
        static char *   CONFIG_QUIT_KEY;
        static char *   CONFIG_SCREEN_SHOT_KEY;
        static char *   CONFIG_MOUSE_INPUT_AXIS;
        static char *   CONFIG_MOUSE_INPUT_THRESHOLD;
        static char *   CONFIG_USE_ANIMATIONS;
        static char *   CONFIG_TURBO_SCROLL_THRESHOLD;
        static char *   CONFIG_LARGE_MENU_THRESHOLD;
        static char *   CONFIG_EFFECT_VOLUME;
        static char *   CONFIG_UNSELECTED_FG_COLOR;
        static char *   CONFIG_UNSELECTED_OUTLINE_COLOR;
        static char *   CONFIG_SELECTED_FG_COLOR;
        static char *   CONFIG_SELECTED_OUTLINE_COLOR;
    };

    class Emu_Menu_Listener
    {
      public:
        virtual void    game_selected(Emulator * emulator, Game * game) = 0;
        virtual void    quit_selected() = 0;
    };

    class Null_Emu_Menu_Listener : public Emu_Menu_Listener
    {
      public:
        void            game_selected(Emulator * emulator, Game * game);
        void            quit_selected();
    };

    class Launcher_Controller : public Emu_Menu_Listener,
                                public Delay_VC_Listener,
                                public Log_Rotate_Listener,
                                public Background_Music_Listener,
                                public Screen_Saver_Listener
    {
      public:
        Launcher_Controller(bool is_first_time) throw (InitError);
        virtual ~Launcher_Controller();

        /**
         * Runs the menu until a game is launched.
         *
         * @return ???
         */
        int launch_game();

        virtual void game_selected(Emulator * emulator, Game * game);
        virtual void quit_selected();
        virtual void delay_vc_expired();
        virtual void rotate_progress(const Log_Rotate_Event * e);

        /**
         * Notification that the skip to next song key was pressed.
         * Delegates this to the background music controller.
         *
         * @param e Event.
         */
        virtual void skip_to_next_song(const Background_Music_Event * e);

        /**
         * Notification that start song over key was pressed.
         * Delegates this to the background music controller.
         *
         * @param e Event.
         */
        virtual void start_song_over(const Background_Music_Event * e);


        /**
         * Notification that the screen saver has changed state.
         */
        virtual void screen_saver_changed_state(const Screen_Saver_Event * e);

        /**
         * Returns the current version, as a string.
         *
         * @return Current version;
         */
        static string get_version_string();

      private:
        typedef vector<Emulator *>      EmuListImpl;
        typedef vector<string>  RomListImpl;
        enum State { INIT, RUNNING, DONE };
        static const int        UPDATE_PRIORITY         = 50;
        static const int        DRAW_PRIORITY           = 50;

        void            restore_state();
        void            save_state();
        void            load_font() throw (InitError);
        void            save_pcx_font();
        void            write_script_header(ostream & file);
        void            write_command(ofstream & file, const string & command,
                                      const string & output_log);

        Engine *        p_engine;
        DataFile *      p_data_file;
        State           p_state;

        FONT *          p_loaded_font;
        string          p_font_file;
        int             p_font_file_size;
        int             p_font_size;
        bool            p_antialias_font;
        bool            p_is_debug_on;

        Config_File *   p_config_file;
        Config_File *   p_state_file;

        Emulator_Manager *      p_emulator_manager;
        Emu_Menu_Controller *   p_emu_menu_controller;
        Background_Music_Controller *   p_bg_controller;
        Delay_VC *              p_delay;
        Splash_VC *             p_splash;
        Background_Music_VC *   p_bg_view;
        Screen_Saver_VC *       p_screen_saver;

        int             p_exit_code;

        static const unsigned   MAX_FILE_NAME         = 256;

        static char *           DATA_FILE;
        static char *           BATCH_FILE;
        static char *           BATCH_LOG_FILE;
        static char *           CONFIG_DIR;
        static char *           FONT_PCX_CACHE;

        static char *           CONFIG_FILE;
        static char *           STATE_FILE;

        static char *           MAIN_SECTION;
        static char *           MAIN_BACKGROUND_MUSIC;
        static char *           MAIN_USE_SOFTWARE_MIXER;
        static char *           MAIN_ANIMATION_METHOD;
        static char *           MAIN_USE_VSYNC;
        static char *           MAIN_USE_VRAM;
        static char *           MAIN_USE_FULLSCREEN;
        static char *           MAIN_MENU_ORIENTATION;
        static char *           MAIN_FONT_FILE;
        static char *           MAIN_FONT_SIZE;
        static char *           MAIN_ANTIALIAS_FONT;
        static char *           MAIN_RESOLUTION;
        static char *           MAIN_DEBUG;
        static char *           MAIN_SCREEN_SAVER_TIME;
        static char *           MAIN_SCREEN_SAVER_TEXT;
        static char *           MAIN_SKIP_SONG_KEY;
        static char *           MAIN_START_SONG_OVER_KEY;

        static const int        MAIN_BACKGROUND_MUSIC_NONE      = 0;
        static const int        MAIN_BACKGROUND_MUSIC_MP3       = 1;
        static const int        MAIN_BACKGROUND_MUSIC_CD        = 2;

        static const int        MAIN_ANIMATION_DOUBLE_BUFFERING = 0;
        static const int        MAIN_ANIMATION_PAGE_FLIPPING    = 1;

        static const int        MAIN_HORIZONTAL_MENU  = 0;
        static const int        MAIN_VERTICAL_MENU    = 1;

        static char *           STATE_SECTION;
        static char *           STATE_LAST_FONT_FILE;
        static char *           STATE_LAST_FONT_FILE_SIZE;
        static char *           STATE_LAST_FONT_SIZE;
        static char *           STATE_LAST_ANTIALIAS_FONT;

        static char *           DEFAULT_FONT_FILE;
        static const int        DEFAULT_HORIZONTAL_WIDTH        = 800;
        static const int        DEFAULT_HORIZONTAL_HEIGHT       = 600;
        static const int        DEFAULT_VERTICAL_WIDTH          = 640;
        static const int        DEFAULT_VERTICAL_HEIGHT         = 480;
    };

    /**
     * Interface for background music controller.  Background music is
     * played as a background task.  This controls the background task
     * to initialize it and keep track of the state of the background
     * task.  State can be saved and restored to a configuration file.
     */
    class Background_Music_Controller
    {
      public:
        virtual ~Background_Music_Controller();

        /**
         * Loads configuration parameters from a configuration file.
         *
         * @param config_file Configuration file to use.
         */
        virtual void load_config(Config_File * config_file) = 0;

        /**
         * Restores state from a configuration file.  State is used to
         * keep track of current song, position in song, etc.
         *
         * @param sate_file Configuration file to use.
         */
        virtual void restore_state(Config_File * state_file) = 0;

        /**
         * Sves state to a configuration file.  This is used so that
         * when a restore is done later, music playback will continue
         * exactly where it left off.
         *
         * @param sate_file Configuration file to use.
         */
        virtual void save_state(Config_File * state_file) = 0;

        /**
         * Stops playing the current song and goes the next song in
         * the queue.
         */
        virtual void skip_to_next_song() = 0;

        /**
         * Starts the current song over from the beginning.
         */
        virtual void start_song_over() = 0;
    };

    /**
     * A background music controller that provides the interface, but
     * does nothing.  This should be used if no background music
     * should be played.
     */
    class Null_Music_Controller : public Background_Music_Controller
    {
      public:
        virtual void load_config(Config_File * config_file);
        virtual void restore_state(Config_File * state_file);
        virtual void save_state(Config_File * state_file);
        virtual void skip_to_next_song();
        virtual void start_song_over();
    };

    /**
     * Controls a background task for playing music off an audio CD.
     */
    class CD_Music_Controller : public Background_Music_Controller
    {
      public:
        CD_Music_Controller(Engine * engine);
        virtual ~CD_Music_Controller();

        virtual void load_config(Config_File * config_file);
        virtual void restore_state(Config_File * config_file);
        virtual void save_state(Config_File * config_file);
        virtual void skip_to_next_song();
        virtual void start_song_over();

      private:
        static char *   CONFIG_SECTION;
        static char *   CONFIG_CD_VOLUME;

        static char *   STATE_SECTION;
        static char *   STATE_LAST_CD_FRAME;
        static char *   STATE_LAST_CD_TOTAL_FRAMES;

        /** The view for this controller. */
        CD_Background * p_cd_background;
    };

    /** 
     * Controls a background task for playing music from an MP3
     * playlist.
     */
    class MP3_Music_Controller : public Background_Music_Controller
    {
      public:
        MP3_Music_Controller(Engine * engine);
        virtual ~MP3_Music_Controller();
        
        virtual void load_config(Config_File * config_file);
        virtual void restore_state(Config_File * config_file);
        virtual void save_state(Config_File * config_file);
        virtual void skip_to_next_song();
        virtual void start_song_over();

      private:
        static char *   CONFIG_SECTION;
        static char *   CONFIG_MP3_PLAYLIST_FILE;
        static char *   CONFIG_MP3_VOLUME;

        static char *   STATE_SECTION;
        static char *   STATE_LAST_MP3_FILE;
        static char *   STATE_LAST_MP3_POSITION;
        static char *   STATE_LAST_MP3_FILE_INDEX;

        /**
         * Parses a playlist from a flat file to a vector of songs.
         *
         * @param playlist_file Playlist in flat file.
         * @param playlist Vector of songs.
         */
        void read_playlist(const string & playlist_file,
                           vector<string> & playlist);

        /** The view for this controller. */
        MP3_Background * p_mp3_background;

        /** The current playlist, as a vector of songs. */
        vector<string>  p_playlist;
    };
}

#endif  /* _CONTROL_H_ */

/* Local Variables: */
/* mode: c++ */
/* End: */
