#ifndef _ENGINE_H_
#define _ENGINE_H_

#include "allegro.h"

#include <string>
#include <map>
#include <list>

using namespace std;

/**
 * GameEngine namespace.
 */
namespace GameEngine
{
    class Drawable;
    class Updatable;
    class Background_Task;
    class Timer;
    class Timer_Factory;
    class Keypress_Watcher;
    class Keypress_Factory;
    class Joystick_Watcher;
    class Joystick_Factory;
    class GEContainer;
    class Engine_Manager;
    class Timer_Manager;
    class Keypress_Manager;
    class Joystick_Manager;

    /**
     * An initialization error.
     */
    class InitError
    {
      public:
        /**
         * Construct a new error.
         *
         * @param error_string String describing the error.
         */
        InitError(const string & error_string);
        
        /**
         * Gets the error message for this exception.
         *
         * @return Error message.
         */
        const string & error_string();

      protected:
        string  p_error_string;
    };

    /**
     * A game engine that abstracts away the details of animation.
     * The user code need not be aware of the different animation
     * methods used.  All optimizations are taken care of.
     */
    class Engine
    {
      public:
        /**
         * All possible animation methods.
         */
        enum Animation_Method
        {
            /**
             * A buffer the same size of the screen.  Each object
             * draws the this buffer.  When the frame is ready to be
             * displayed, it is blit to the screen.  For smooth
             * animation, the final blit should wait for a vertical
             * sync so that the blit can take place during the
             * refresh.  This method can be slowest since the whole
             * screen is blit for each game sycle.
             */
            DOUBLE_BUFFERING,

            /**
             * The screen is split into two frames, one active
             * (currently displayed) and the other inactive.  Each
             * object draws to the inactive frame.  When the frame is
             * ready to be displayed, the pages are instantaneously
             * flipped (the inactive becomes active and vice versa).
             * This method is faster than double buffering, but the
             * video hardware and driver must support page flipping.
             */
            PAGE_FLIPPING,

            // Some day add TRIPLE_BUFFERING
        };

        /**
         * Returns the engine singleton, creating a new one (and
         * initializing Allegro) if necessary.
         *
         * @return Singleton instance of the engine.
         */
        static Engine * get_instance();

        /**
         * Cleans up internal objets.
         */
        ~Engine();

        /**
         * Initializes the sound subsystem.  This should autodetect
         * the sound driver.  On Windows, this is a DirectSound driver
         * with a hardware mixer.  Sometimes this DirectSound mixer is
         * buggy, thus a DirectSound software mixer can be forced.
         *
         * @param use_software_mixer Force a software mixer (Windows only)
         */
        void initialize_sound(bool use_software_mixer)
            throw (InitError);

        /**
         * Sets the graphics mode.  This should be called once and
         * only once before running the engine.  Only true color video
         * modes are supported (i.e. no palettes).  If the color depth
         * chosen is 16 BPP and that cannot be set, 15 BPP will also
         * be tried automatically.
         *
         * @param width Width of the screen resolution.
         * @param height Height of the screen resolution.
         * @param color_depth The bits per pixel.
         * @param use_fullscreen.
         * @throws InitError If the mode could not be set as specified.
         */
        void set_graphics_mode(int width, int height, int color_depth,
                               bool use_fullscreen)
            throw (InitError);

        /**
         * Sets the animation mode of the the engine.
         */
        void set_animation_mode(Animation_Method animation_method,
                                bool use_vram, bool use_vsync)
            throw (InitError);

        /**
         * Enters the main loop.  This function will not return until
         * this engine has been stopped.
         */
        void run();

        /**
         * Terminates the main loop causing run() to return.
         */
        void stop();

        unsigned int    updates_per_second();

        /**
         * Gets the current frames per seconds as calculated for the
         * previous second.
         *
         * @return Current frames per second.
         */
        unsigned int get_frames_per_second();

        /**
         * Gets the overall average frames per second for the whole
         * lifetime of this engine.
         *
         * @return Average frames per second.
         */
        float get_average_frames_per_second();

        int             get_catchup();
        Animation_Method animation_method();
        bool            use_vram();
        bool            use_vsync();
        unsigned int    elapsed_seconds();

        GEContainer *   get_root_container();
        Timer_Factory * get_timer_factory();
        Keypress_Factory * get_keypress_factory();
        Joystick_Factory * get_joystick_factory();

        /**
         * Sets the background task.
         *
         * @param bg_task New background taask.
         */
        void set_background_task(Background_Task * bg_task);

        /**
         * Polls the background task.  Normally this is done
         * automatically in the main loop.  If you are doing an
         * operation that is going to take a lot of time, you may want
         * to manualy call this method periodically to make sure that
         * the background task doesn't get starved.
         */
        void poll_background_task();

        /**
         * Forces the game cycles to catchup without actually running
         * through each remaining cycle.  This is useful if a long
         * operation is performed and you know that nothing has been
         * updated.  This is better than delaying for an indeterminant
         * amount of time, though it is still a hack.
         */
        void catchup();

        /**
         * Gets the X-coordinate of the infinte mouse location.  This
         * coordinate never hits the edge of the screen as the mouse
         * moves, but keeps going on forever.  Of course it is not
         * really infinite.  It may roll over, if not set back within
         * range occasionally.
         *
         * @return X-coordinate of infinite mouse location.
         */
        int get_infinite_mouse_x();
        
        
        /**
         * Gets the Y-coordinate of the infinite mouse location.  This
         * coordinate never hits the edge of the screen as the mouse
         * moves, but keeps going on forever.  Of course it is not
         * really infinite.  It may roll over, if not set.
         *
         * @return Y-coordinate of infinite mouse location.
         */
        int get_infinite_mouse_y();

        /**
         * Sets the X-coordinate of the infnite mouse location.
         *
         * @param infinite_mouse_x New X-coordinate for infinite mouse
         * location.
         */
        void set_infinite_mouse_x(int infinite_mouse_x);

        /**
         * Sets the Y-coordinate of the infinite mouse location.
         *
         * @param infinite_mouse_y New Y-cooridinate for infinite
         * mouse location.
         */
        void set_infinite_mouse_y(int infinite_mouse_y);

        /**
         * Gets the number of seconds that the system has been idle.
         * An idle system is one that has not keypresses or mouse
         * movement.
         *
         * @return The number of seconds the system has been idle.
         */
        unsigned int get_idle_timer();

        /**
         * Resets the number of seconds the system has been idle to
         * zero.  The engine automatically detects keyboard and mouse
         * activity.  This should only be called if other input
         * activity is detected (like joysticks).
         */
        void reset_idle_timer();
        

      private:
        /** The frames per second that the engine is trying to achieve. */
        static const int        TARGET_FPS      = 60;

        /** The number of frame bitmaps to create. */
        static const int        NUM_FRAMES      = 2;

        /**
         * Update priority of timer manager.  This should always be
         * the highest priority.
         */
        static const int        TIMER_MANAGER_UPDATE_PRIORITY       = 0;

        /** Update priority of keypress manager. */
        static const int        KEYPRESS_MANAGER_UPDATE_PRIORITY    = 1;

        /** Priority of joystick manager. */
        static const int        JOYSTICK_MANAGER_UPDATE_PRIORITY    = 2;

        /**
         * Update p of engine manager.  This should alwyas be after
         * the other managers as it may use their featurs.
         */
        static const int        ENGINE_MANAGER_UPDATE_PRIORITY      = 10;


        /**
         * Draw priorty of engine manager.  This should always be
         * after the other managers.
         */
        static const int        ENGINE_MANAGER_DRAW_PRIORITY        = 10;

        /**
         * Initializes Allegro and some internal objects, but cannot
         * be run right away.  The graphics and animation modes must
         * be set prior to running the engine.
         *
         * @throws InitError If Allegro could not be initialized.
         */
        Engine()
            throw (InitError);

        void            initialize(Animation_Method method, bool use_vram,
                                   bool use_vsync) throw (InitError);
        void            draw_one_frame();
        void            do_one_cycle();
        void            get_keyboard_input();
        void            log_os_type();
        void            log_digi_card();

        /** The singleton instance. */
        static Engine * s_instance;

        /**
         * Container for all system level objects. System level
         * objects are updated before user objects and drawn after
         * user objects.
         */
        GEContainer *   p_system_container;
        
        /** Container for all user level objects. */
        GEContainer *   p_root_container;

        Background_Task *       p_background_task;
        Background_Task *       p_null_background_task;

        BITMAP *        p_frame;
        BITMAP *        p_frames[NUM_FRAMES];
        bool            p_end_game;
        bool            p_paused;

        Animation_Method        p_animation_method;
        bool            p_use_vram;
        bool            p_use_vsync;

        /** If the mouse has been installed, set to true. */
        bool            p_is_mouse_installed;

        /** The X-coordinate of the infinite mouse location. */
        int             p_infinite_mouse_x;

        /** The Y-coordinate of the infinite mouse locaiton. */
        int             p_infinite_mouse_y;

        int             p_current_frame;
        unsigned int    p_actual_cycle;
        unsigned int    p_loop_count;
        unsigned int    p_catchup;

        /** The engine manager. */
        Engine_Manager *    p_engine_manager;

        /** The timer manager. */
        Timer_Manager *     p_timer_manager;

        /** The keypress manager. */
        Keypress_Manager *  p_keypress_manager;

        /** The joystick manager. */
        Joystick_Manager *  p_joystick_manager;
    };

    inline Engine::Animation_Method Engine::animation_method()
        { return p_animation_method; }
    
    inline bool Engine::use_vram()
        { return p_use_vram; }

    inline bool Engine::use_vsync()
        { return p_use_vsync; }

    class Drawable
    {
      public:
        virtual         ~Drawable() { }
        virtual void    draw(BITMAP * frame) = 0;
    };

    class Updatable
    {
      public:
        virtual         ~Updatable() { }
        virtual bool    update() = 0;
    };

    class Background_Task
    {
      public:
        virtual         ~Background_Task() { }
        virtual void    poll() = 0;
        virtual void    resume() = 0;
        virtual void    suspend() = 0;
    };

    class Null_Background_Task : public Background_Task
    {
      public:
        void            poll();
        void            resume();
        void            suspend();
    };

    class Timer
    {
      public:
        virtual         ~Timer() { }
        virtual void    set(unsigned int milliseconds) = 0;
        virtual void    set_interval(unsigned int milliseconds) = 0;
        virtual void    disable() = 0;
        virtual bool    expired() = 0;
        virtual bool    stopped() = 0;
    };

    class Timer_Factory
    {
      public:
        virtual         ~Timer_Factory() { }
        virtual Timer * create_timer() = 0;
    };

    class Keypress_Watcher
    {
      public:
        virtual         ~Keypress_Watcher() { }
        virtual void    set_scancode(int scancode) = 0;
        virtual bool    is_pressed() = 0;
        virtual bool    is_artificial() = 0;
        virtual bool    is_physically_pressed() = 0;
        virtual void    set_repeat(int initial_delay, int repeat_delay) = 0;
        virtual void    set_continuous() = 0;
        virtual void    set_one_time() = 0;
    };

    class Keypress_Factory
    {
      public:
        virtual Keypress_Watcher * create_keypress_watcher(int scancode) = 0;
    };
     
    typedef enum { BUTTON_A, BUTTON_B, JOY_UP, JOY_DOWN, JOY_LEFT,
                   JOY_RIGHT } Joystick_Codes;

    class Joystick_Event
    {
      public:
        virtual         ~Joystick_Event() { }
        virtual bool    is_triggered() = 0;
    };

    class Joystick_Button_Event : public Joystick_Event
    {
      public:
        Joystick_Button_Event(int joystick, int button);

        bool        is_triggered();

      protected:
        int         p_joystick;
        int         p_button;
    };

    class Joystick_Digital_Axis_Event : public Joystick_Event
    {
      public:
        typedef enum {POSITION_ONE, POSITION_TWO} Position;
        Joystick_Digital_Axis_Event(int joystick, int stick, int axis,
                                    Position position);
        bool        is_triggered();

      protected:
        int         p_joystick;
        int         p_stick;
        int         p_axis;
        Position    p_position;
    };

    class Joystick_Watcher
    {
      public:
        virtual         ~Joystick_Watcher() { }
        virtual void    set_joystick_event(Joystick_Event * event) = 0;
        virtual bool    is_pressed() = 0;
        virtual void    set_repeat(int initial_delay, int repeat_delay) = 0;
        virtual void    set_continuous() = 0;
        virtual void    set_one_time() = 0;
    };

    class Joystick_Factory
    {
      public:
        virtual Joystick_Watcher * create_joystick_watcher(
            Joystick_Event * joystick_event) = 0;

        virtual bool is_joystick_detected() = 0;
    };

    class GEContainer: public Drawable, public Updatable
    {
      public:
        GEContainer();
        virtual         ~GEContainer();
        virtual void    draw(BITMAP * bitmap);
        virtual bool    update();

        void            add_drawable(Drawable * drawable, int draw_priority);
        void            remove_drawable(Drawable * drawable);

        void            add_updatable(Updatable * updatable,
                                      int update_priority);
        void            remove_updatable(Updatable * updatable);

      private:
        typedef multimap<int, Drawable *>        Drawable_List;
        typedef multimap<int, Updatable *>       Updatable_List;

        Drawable_List   p_drawable_objects;
        Updatable_List  p_updatable_objects;
    };

};
#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
