#ifndef _GLAUNCH_H_
#define _GLAUNCH_H_

#include <string>
#include <vector>
#include <map>
#include <functional>
#include "allegro.h"
#include "glch_dat.h"
#include "engine.h"

class       Config_File;

namespace GameLauncher
{
    using namespace GameEngine;

    class       Launcher;
    class       Emulator_Manager;
    class       Emulator_Manager_Listener;
    class       Emulator;
    class       Emulator_Listener;
    class       EmulatorConfig;
    class       Game;
    class       DataFile;
    class       Screen_Shot;
    class       Menu;

    class       Emu_Menu_Controller;

    typedef vector<Game *>      Game_List;
    typedef vector<Emulator *>  Emulator_List;
    typedef map<string, string> NameMap;

    struct nocase_less : public binary_function<string, string, bool>
    {
        bool operator()(const string & s1, const string & s2);
    };
    
    class Emulator_Manager_Listener
    {
      public:
        virtual ~Emulator_Manager_Listener();
        virtual void    emulator_manager_changed() = 0;
        virtual void    poll_emulator_manager() = 0;
    };

    class Emulator_Listener
    {
      public:
        virtual ~Emulator_Listener();
        virtual void    poll_emulator_listener() = 0;
    };

    class Emulator_Manager : public Emulator_Listener
    {
      public:
        Emulator_Manager(Config_File * config_file, Config_File * state_file);
        ~Emulator_Manager();

        void            restore_state();
        void            save_state();
        void            set_listener(Emulator_Manager_Listener * listener);

        unsigned        get_number_of_emulators();
        Emulator *      get_current_emulator();
        unsigned        get_current_emulator_index();
        void            increment_emulator_index();
        void            decrement_emulator_index();

        // Emulator_Listener interface
        void            poll_emulator_listener();

      private:
        static char *   CONFIG_SECTION;
        static char *   CONFIG_EMULATORS;
        static char *   CONFIG_PATH_PREFIX;

        static char *   STATE_SECTION;
        static char *   STATE_CURRENT_EMU;

        void            load_emulator_if_necessary();

        Emulator_List   p_emulator_list;
        vector<string> *        p_emulator_names;
        vector<bool>    p_emulator_loaded;
        unsigned        p_current_emulator;
        string          p_path_prefix;
        Config_File *   p_state_file;

        Emulator_Manager_Listener *     p_listener;
    };

    class Null_Emulator_Manager_Listener : public Emulator_Manager_Listener
    {
      public:
        static Emulator_Manager_Listener * instance();
        void            emulator_manager_changed();
        void            poll_emulator_manager();

      private:
        static Emulator_Manager_Listener *      p_instance;
    };

    class InvalidEmulatorConfig { };

    class Emulator
    {
      public:
        Emulator(const string & config_name,
                 const string & path_prefix,
                 Emulator_Listener * listener = 0
                 ) throw (InvalidEmulatorConfig);
        virtual ~Emulator();

        int             number_of_games() const;
        void            get_game_list(Game_List & game_list) const;

        // State related
        unsigned        get_selected_index() const;
        unsigned        get_first_shown_index() const;
        void            set_selected_index(unsigned int selected_index_);
        void            set_first_shown_index(unsigned int first_shown_index_);

        void            restore_state(Config_File * state_file);
        void            save_state(Config_File * state_file);

        // Configuration parameters
        const string &  config_directory() const;
        const string &  config_name() const;
        const string &  name() const;
        const string &  description() const;
        const string &  directory() const;
        const string &  executable() const;
        const string &  rom_directory() const;
        const string &  global_options() const;
        const string &  screen_shot_directory() const;
        const string &  map_file() const;
        bool            options_before_rom() const;
        bool            use_rom_file() const;
        bool            use_extension() const;
        bool            use_rom_directory() const;
        const string &  get_pre_command() const;
        const string &  get_post_command() const;
        string          get_screen_shot_file_name(Game * game);
        string          get_command_to_run(Game * game);


      protected:
        Emulator();
        typedef map<string, Game *, nocase_less> GameListImpl;

        void            load_config();
        void            load_name_map();
        void            add_game(const string & rom_base,
                                 const string & extension);
        void            add_games_from_rom_file();
        void            search_for_games();
        string          append_prefix(const string & path);
        bool            is_valid_extension(const string & extension);

        static char *   CONFIG_OPTIONS_SECTION;

        NameMap         p_name_map;
        GameListImpl    p_games;

        string          p_config_directory;
        string          p_config_name;
        Config_File *   p_config_file;

        string          p_path_prefix;
        string          p_name;
        string          p_description;
        string          p_directory;
        string          p_pre_command;
        string          p_executable;
        string          p_post_command;
        string          p_rom_directory;
        vector<string> * p_rom_extension;
        string          p_global_options;
        string          p_screen_shot_directory;
        string          p_map_file;
        bool            p_options_before_rom;
        bool            p_use_rom_file;
        bool            p_use_extension;
        bool            p_use_rom_directory;
        unsigned int    p_selected_index;
        unsigned int    p_first_shown_index;

        Emulator_Listener *     p_emulator_listener;

        static char *   CONFIG_SECTION;
        static char *   CONFIG_NAME;
        static char *   CONFIG_DESCRIPTION;
        static char *   CONFIG_DIRECTORY;
        static char *   CONFIG_EXECUTABLE;
        static char *   CONFIG_USE_ROM_FILE;
        static char *   CONFIG_ROM_EXTENSION;
        static char *   CONFIG_ROM_DIRECTORY;
        static char *   CONFIG_USE_EXTENSION;
        static char *   CONFIG_USE_ROM_DIRECTORY;
        static char *   CONFIG_GLOBAL_OPTIONS;
        static char *   CONFIG_OPTIONS_BEFORE_ROM;
        static char *   CONFIG_MAP_FILE;
        static char *   CONFIG_SCREEN_SHOT_DIRECTORY;
        static char *   CONFIG_PRE_COMMAND;
        static char *   CONFIG_POST_COMMAND;

        static char *   CONFIG_ROM_TYPE_INTERNAL;
        static char *   CONFIG_ROM_TYPE_EXTERNAL;
        static char *   CONFIG_YES;
        static char *   CONFIG_NO;

        static char *   STATE_SECTION;
        static char *   STATE_SELECTED_INDEX;
        static char *   STATE_FIRST_SHOWN_INDEX;
    };

    class InvalidEmulator : public Emulator
    {
      public:
        InvalidEmulator(const string & configName,
                        Emulator_Listener * listener = 0);
    };

    class Null_Emulator_Listener : public Emulator_Listener
    {
      public:
        static Emulator_Listener * instance();
        void    poll_emulator_listener();

      private:
        static Emulator_Listener *      p_instance;
    };

    class Game
    {
      public:
        Game(const string & full_name, const string & rom_name,
             const string & extension);
        
        string          get_full_name() const;
        string          get_rom_name() const;
        string          get_extension() const;

      private:
        string  p_full_name;
        string  p_rom_name;
        string  p_extension;
    };

    class DataFile
    {
      public:
        DataFile(DATAFILE * df);

        FONT *          list_font() const;
        FONT *          fontb() const;

        SAMPLE *        selection_moved() const;
        SAMPLE *        change_game() const;
        SAMPLE *        quit() const;
        SAMPLE *        start_game() const;

        BITMAP *        screen_shot_icon() const;

      private:
        DATAFILE *      p_df;

    };

    inline DataFile::DataFile(DATAFILE * df) {
        p_df = df;
    }

    inline FONT * DataFile::list_font() const {
        return (FONT *) p_df[LIST_FONT].dat;
    }

    inline FONT * DataFile::fontb() const {
        return (FONT *) p_df[FONTB].dat;
    }

    inline SAMPLE * DataFile::selection_moved() const {
        return (SAMPLE *) p_df[SELECTION_MOVED].dat;
    }

    inline SAMPLE * DataFile::change_game() const {
        return (SAMPLE *) p_df[CHANGE_GAME].dat;
    }

    inline SAMPLE * DataFile::quit() const {
        return (SAMPLE *) p_df[QUIT].dat;
    }

    inline SAMPLE * DataFile::start_game() const {
        return (SAMPLE *) p_df[START_GAME].dat;
    }

    inline BITMAP * DataFile::screen_shot_icon() const {
        return (BITMAP *) p_df[SSHOT_ICON].dat;
    }
};


#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
