#ifndef _GUI_LIST_H_
#define _GUI_LIST_H_

#include <list>
#include "engine.h"
#include "gui/comp.h"
#include "gui/util.h"

namespace GUI
{
    class Icon;
    class Label;
    class List_UI;

    using GameEngine::GEContainer;
    using GameEngine::Updatable;
    using GameEngine::Drawable;
    
    /**
     * The interface an object must implement to be viewed in a List
     * component.  This interface specifies random access to the list
     * items.
     */
    class List_Model
    {
      public:
        /**
         * Virtual destructor.
         */
        virtual ~List_Model();

        /**
         * Gets the text for a list item.
         *
         * @param index Index of list item.
         * @return Text of list itemm.
         */
        virtual string get_item_text_at(unsigned index) const = 0;

        /**
         * Gets the icon for a list item.
         *
         * @param index Index of list item.
         * @return Icon of list item.
         */
        virtual Icon * get_item_icon_at(unsigned index) const = 0;

        /**
         * Gets the number of items in this list.
         *
         * @return The number of items in this list.
         */
        virtual unsigned get_size() const = 0;
    };

    /**
     * A scrollable list of items.  Each item contains text and an
     * icon.  One item is always selected and may be displayed with
     * different colors.  The selected item may be moved a line at a
     * time or a page at a time.  As items are scrolled on and off the
     * screen, smooth scrolling may be used.
     */
    class List : public Component
    {
      public:
        /** The base number of millseconds to delay. */
        static const int        BASE_SCROLL_DELAY           = 100;

        enum Turbo_Options
        {
            RESET_TURBO_COUNT,
            INCREMENT_TURBO_COUNT
        };

        /**
         * Construct a new list with the given bounding box and an
         * empty model.
         *
         * @param x X coordinate of upper left corner.
         * @param y Y coordinate of upper left corner.
         * @param width Width.
         * @param height Height.
         */
        List(int x, int y, int width, int height);

        /**
         * Construct a new list with the given bounding box and an empty
         * model.
         *
         * @param bounding_box
         */
        List(const Rectangle & bounding_box);

        /**
         * Destructor.
         */
        ~List();

        /**
         * Gets the x coordinate of the upper left corner.
         *
         * @return X coordinate of the upper left corner.
         */
        int get_x() const;

        /**
         * Gets the y coordinate of the upper left corner.
         *
         * @return Y coordinate of the upper left corner.
         */
        int get_y() const;
        
        /**
         * Gets the width of this list.
         *
         * @return The width of this list.
         */
        int get_width() const;

        /**
         * Gets the height of this list.
         *
         * @return The height of this list.
         */
        int get_height() const;
        
        /**
         * Sets the model for this list.
         *
         * @param model The new model.
         */
        void set_model(const List_Model * model);
        
        /**
         * Gets the model for this list.
         *
         * @return The model for this list.
         */
        const List_Model * get_model() const;

        /**
         * Sets the text justification for all items in this list.
         * The default is centered justification.
         *
         * @param justification The justification of the item text.
         */
        void set_justification(Justification justification);

        /**
         * Gets the text justification for all items in this list.
         *
         * @return The justification of the item text.
         */
        Justification get_justification() const;
        
        /**
         * Sets the position of the selection.  The first shown index
         * is the first item to be visible in the list.  This allows
         * the selection to be set to any item displayed.
         *
         * @param first_shown_index Index of the first displayed item.
         * @param selected_index Index of selected item.
         */
        void set_selection_position(unsigned first_shown_index,
                                    unsigned selected_index);
        
        /**
         * Gets the index of the selected item.
         *
         * @return Index of the selected item.
         */
        unsigned get_selected_index() const;

        /**
         * Gets the index of the first item visible in this list.
         *
         * @return Index of the first shown item.
         */
        unsigned get_first_shown_index() const;
        
        /**
         * Sets the colors to use for the selected item.
         *
         * @param text Text color.
         * @param outline Outline color.
         */
        void set_selected_colors(int text, int outline);

        /**
         * Gets the text color to be used for the selected item.
         *
         * @return Text color for the selected item.
         */
        int get_selected_text_color() const;

        /**
         * Gets the outline color to be used for the selected item.
         *
         * @return Outline color for the selected item.
         */
        int get_selected_outline_color() const;
        
        /**
         * Sets the color to use for unselected items.
         *
         * @param text Text color.
         * @param outline Outline color.
         */
        void set_unselected_colors(int text, int outline);

        /**
         * Gets the text color to be used for unselected items.
         *
         * @return Text color for unselected items.
         */
        int get_unselected_text_color() const;

        /**
         * Gets the outline color to be used for unselected items.
         *
         * @return Outline color for unselected items.
         */
        int get_unselected_outline_color() const;
        
        /**
         * Move the selection up one line.  If the selection is at the
         * top of the displayed items, all items are scrolled down by
         * one.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void line_up(Scroll_Options scroll_options);

        /**
         * Move the selection down one line.  If the selection is at
         * the bottom of the displayed items, all items are scrolled
         * up by one.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void line_down(Scroll_Options scroll_options);

        /**
         * Move the selection up one page.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void page_up(Scroll_Options scroll_options);

        /**
         * Move the selection down one page.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void page_down(Scroll_Options scroll_options);

        /**
         * Hide all items with half going to the left and half going
         * to the right.
         */
        void hide_split();

        /**
         * Hide all items to the left.
         */
        void hide_left();

        /**
         * Hide all items to the right.
         */
        void hide_right();
        
        /**
         * Hide all items to the top;
         */
        void hide_top(int milliseconds);

        /**
         * The same as hide_split() but the selected item stays
         * displayed and is centered within the list geometry.
         *
         * @param milliseconds.
         */
        void hide_split_center_selected(int milliseconds);

        /**
         * Unhides all items.
         */
        void unhide();

        /**
         * Updates the position of all items.
         */
        virtual bool update();

        /**
         * Draws all items at their current location.
         *
         * @param frame Active screen frame.
         */
        virtual void draw(BITMAP * frame);

      private:
        /**
         * Moves the selection up, specifying how many items the
         * selection is moved by and how many items to scroll by.
         *
         * @param move_up The selected item is moved by this many
         * items.
         *
         * @param scroll_up The displayed list is scrolled by this
         * many items.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void move_selection_up(unsigned move_up, unsigned scroll_up,
                               Scroll_Options scroll_options);

        /**
         * Moves the selection down, specifying how many items the
         * selection is moved by and how many items to scroll by.
         *
         * @param move_down The selected item is moved by this many
         * items.
         *
         * @param scroll_down The displayed list is scrolled by this
         * many items.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void move_selection_down(unsigned move_down, unsigned scroll_down,
                                 Scroll_Options scroll_options);

        /**
         * Gets the index of the last item visible in this list.
         */
        unsigned get_last_shown_index() const;

        /**
         * Calculates the base scroll delay from the scroll options.
         *
         * @param scroll_options Options to use when calculating delay.
         * @return The base scroll delay.
         */
        int get_base_scroll_delay(Scroll_Options scroll_options) const;

        /** The UI for this list. */
        List_UI * p_ui;

        /** The x coordinate of the upper left corner. */
        int p_x;

        /** The y coordinate of the upper left corner. */
        int p_y;

        /** The width, in pixels */
        int p_width;

        /** The height, in pixels. */
        int p_height;

        /** The model. */
        const List_Model * p_model;

        /**
         * Set to true if the list has changed in some visible way
         * that the UI would need to know about.
         */
        bool p_list_changed;

        /** The text color for the selected item. */
        int p_selected_text_color;

        /** The outline color for the selected item. */
        int p_selected_outline_color;

        /** The text color for the unselected items. */
        int p_unselected_text_color;

        /** The outline color for the unselected items. */
        int p_unselected_outline_color;

        /** The index of the selected item. */
        unsigned p_selected_index;

        /** Index of the first visible item. */
        unsigned p_first_shown_index;

        /** The number of items that are visible at one time. */
        unsigned p_num_visible_items;

        /** The justification of the items in the list. */
        Justification p_justification;
    };

    /**
     * The interface that draws the list.  Different implementations
     * have complete control as to how the list may be viewed.  The
     * only restriction as that only the rectangle as specified by the
     * list may be drawn to.
     */
    class List_UI : public Updatable, public Drawable
    {
      public:
        /**
         * Virtual destructor.
         */
        virtual ~List_UI();

        /**
         * Called when a list is attached to a UI.
         */
        virtual void install_list(List * list) = 0;

        /**
         * Gets the number of visible items.
         *
         * @return The number of visible items.
         */
        virtual unsigned get_number_of_visible_items() = 0;
        
        /**
         * Move the selection up relative from where it currently is
         * using smooth scrolling.  The time it takes to scroll is a
         * factor of the base scroll delay.  So to disable smooth
         * scrolling the base scroll delay should be set to 0.
         *
         * @param moved The number of items the selection moved by.
         * @param scroll The number of items that should scroll off
         * the screen.
         * @param base_scroll_delay The minimum time to scroll, in ms.
         */
        virtual void move_selection_up(unsigned moved, unsigned scroll,
                                       int base_scroll_delay) = 0;

        /**
         * Move the selection down relative from where it currently
         * is.  The time it takes to scroll is a factor of the base
         * scroll delay.  So to disable smooth scrolling the base
         * scroll delay should be set to 0.
         *
         * @param moved The number of items the selection moved by.
         * @param scroll The number of items that should scroll off
         * the screen.
         * @param base_scroll_delay The minimum time to scroll, in ms.
         */
        virtual void move_selection_down(unsigned moved, unsigned scroll,
                                         int base_scroll_delay) = 0;

        /**
         * The list has changed and the UI must be redrawn.
         */
        virtual void list_changed() = 0;

        /**
         * All items are hidden with half going to the left and half
         * going to the right.
         */
        virtual void hide_split(int milliseconds) = 0;

        /**
         * Hide all items to the left.
         */
        virtual void hide_left(int milliseconds) = 0;

        /**
         * Hide all items to the right.
         */
        virtual void hide_right(int milliseconds) = 0;
        
        /**
         * Hide all items to the top.
         */
        virtual void hide_top(int milliseconds) = 0;

        /**
         * Unhide all items.
         */
        virtual void unhide(int milliseconds) = 0;

        /**
         * Returns true if the list is currently unhidden.  Note: this
         * should return false if the list moving to become unhidden
         * as well as hidden.
         */
        virtual bool is_unhidden() const = 0;

        /**
         * The same as hide_split(int) but the selected item stays
         * displayed and is centered within the list geometry.
         */
        virtual void hide_split_center_selected(int milliseconds) = 0;
        
        /**
         * Returns true if all items are stopped.
         *
         * @return True if all items are stopped.
         */
        virtual bool are_all_items_stopped() const = 0;

        /**
         * Updates the position of any internal components.
         */
        virtual bool update() = 0;

        /**
         * Draws all internal components at their current location.
         *
         * @param frame Active screen frame.
         */
        virtual void draw(BITMAP * frame) = 0;
    };

    /**
     * A base implementation using lists of Labels to represent the
     * items.  All movement and control of the labels is implemented,
     * however, the display of each item is left unimplemented.  This
     * allows base classes to customize how the labels are displayed.
     * Two lists of labels are used for displaying the labels.  The
     * first list contains the items that are currently visible and
     * the second list contains items that are not visible, called
     * offscreen items.  These offscreen items are used when
     * scrolling.  The number of labels in the offscreen list must be
     * equal to the number of items in the visible list.  If the list
     * is scrolled by a page, all currently visible items will be
     * offscreen after the scroll and all currently offscreen items
     * will be visible after the scroll.  A GEContainer is used to
     * display and draw all labels.
     */
    class Abstract_List_UI : public List_UI
    {
      public:
        /**
         * Inintializes all data structres used by this base
         * implementation.
         */
        Abstract_List_UI();

        /**
         * Destroys all labels in the visible and offscreen lists.
         */
        virtual ~Abstract_List_UI();

        /**
         * Constructs the labels and adds them into the visible and
         * offscreen lists.  Equal number of labels should be created
         * for each list.  Each label should also be added to the
         * container.  This must be implemented by the base class
         * since different labels may be used.
         *
         * @param list List that this UI is attached to.
         */
        virtual void install_list(List * list) = 0;
        
        virtual unsigned get_number_of_visible_items();

        virtual void move_selection_up(unsigned moved, unsigned scroll,
                                       int base_scroll_delay);

        virtual void move_selection_down(unsigned moved, unsigned scroll,
                                         int base_scroll_delay);

        virtual void list_changed();
        
        virtual void hide_split(int milliseconds);

        virtual void hide_left(int milliseconds);

        virtual void hide_right(int milliseconds);
        
        virtual void hide_top(int milliseconds);

        virtual void hide_split_center_selected(int milliseconds);
        
        virtual void unhide(int milliseconds);

        virtual bool is_unhidden() const;

        virtual bool are_all_items_stopped() const;

        /**
         * Updates the container as well as performs any state
         * transitions for the state machine.
         */
        virtual bool update();

        virtual void draw(BITMAP * frame);

      protected:
        /**
         * All possible states for the state machine.
         */
        enum State
        {
            /** The labels are unhidden. */
            UNHIDDEN,
            /** The labels are moving towards UNHIDDEN. */
            UNHIDING,
            /** The labels are scrolling. */
            SCROLLING,
            /** The labels are hidden to the right. */
            HIDDEN_RIGHT,
            /** The labels are moving towards HIDDEN_RIGHT. */
            HIDING_RIGHT,
            /** The labels are hidden to the left. */
            HIDDEN_LEFT,
            /** The labels are moving twoards HIDDEN_LEFT. */
            HIDING_LEFT,
            /** The labels are hidden to the top. */
            HIDDEN_TOP,
            /** The labels are moving towards HIDDEN_TOP. */
            HIDING_TOP,
            /** The labels are hidden half left and half right. */
            HIDDEN_SPLIT,
            /** The labels are moving towards HIDDEN_SPLIT. */
            HIDING_SPLIT,
        };

        /** A linked-list of labels. */
        typedef list<Label *> Label_List;
        
        /** The draw priority of the labels within the container. */
        static const int        DRAW_PRIORITY           = 5;

        /** The update priority of the labels within the container. */
        static const int        UPDATE_PRIORITY         = 5;

        /**
         * Hide all offscreen items somewhere where they cannot be
         * seen.
         */
        void hide_offscreen_items();

        /**
         * Change the colors for the selected item to the unselected
         * colors.
         */
        void unselect_selected();

        /**
         * Change the colors of the selected item to the selected
         * colors.
         */
        void select_selected();

        /**
         * Hides one item to the left over time.  The <tt>index</tt>
         * is relative to the first shown index, so 0 is the first
         * item displayed on the screen.  <tt>index</tt> may also be
         * negative indicating that this item can be moved "above" the
         * visibile items.
         *
         * @param item Item to hide.
         * @param index Index relative to first item displayed.
         * @param milliseconds Time to move this item.
         */
        virtual void hide_item_left(Label * item, int index,
                                    int milliseconds) = 0;

        /**
         * Hides one item to the right over time.  The <tt>index</tt>
         * is relative to the first shown index, so 0 is the first
         * item displayed on the screen.  <tt>index</tt> may also be
         * negative indicating that this item can be moved "above" the
         * visibile items.
         *
         * @param item Item to hide.
         * @param index Index relative to first item displayed.
         * @param milliseconds Time to move this item.
         */
        virtual void hide_item_right(Label * item, int index,
                                     int milliseconds) = 0;

        /**
         * Hides one item to the top over time.  The <tt>index</tt>
         * is relative to the first shown index, so 0 is the first
         * item displayed on the screen.  <tt>index</tt> may also be
         * negative indicating that this item can be moved "above" the
         * visibile items.
         *
         * @param item Item to hide.
         * @param index Index relative to first item displayed.
         * @param milliseconds Time to move this item.
         */
        virtual void hide_item_top(Label * item, int index,
                                   int milliseconds) = 0;

        /**
         * Unhides one item over time.  The <tt>index</tt> is relative
         * to the first shown index, so 0 is the first item displayed
         * on the screen.  <tt>index</tt> may also be negative
         * indicating that this item can be moved "above" the visibile
         * items.
         *
         * @param item Item to unhide.
         * @param index Index relative to first item displayed.
         * @param milliseconds Time to move this item.
         */
        virtual void unhide_item(Label * item, int index,
                                 int milliseconds) = 0;

        /**
         * Hides one item to the left over time.  The <tt>index</tt>
         * is relative to the first shown index, so 0 is the first
         * item displayed on the screen.  <tt>index</tt> may also be
         * negative indicating that this item can be moved "above" the
         * visibile items.
         *
         * @param item Item to hide.
         * @param milliseconds Time to move this item.
         */
        virtual void center_item(Label * item, int milliseconds) = 0;

        /**
         * Hides one offscreen item immediately.
         *
         * @param item Item to hide.
         */
        virtual void hide_offscreen_item(Label * item) = 0;

        /** The parent List. */
        List * p_list;

        /**
         * The number of visible items.  Both p_visible_items and
         * p_offscreen_items should be this size.
         */
        unsigned p_num_visible_items;

        /** List of visible items. */
        Label_List p_visible_items;

        /**
         * List of offscreen items.  Should be the same size as
         * p_visible_items.
         */
        Label_List p_offscreen_items;

        /** Points to the current selected item. */
        Label_List::iterator p_selected_item;

        /**
         * Holds all visible and offscreen labels to simplify drawing
         * and updating them.
         */
        GEContainer * p_container;

        /**
         * The height, in pixels, of one item.
         */
        int p_item_height;

        /** The current state. */
        State p_state;

        /** The justification of the text. */
        Justification p_justification;

        /** The x-coordinate of the upper left corner. */
        int p_x;

        /** The y-coordinate of the upper left corner. */
        int p_y;

        /** The width of the parent list. */
        int p_width;

        /** The height of the parent list. */
        int p_height;
    };
};

#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
