#ifndef _GUI_UTIL_H_
#define _GUI_UTIL_H_

#include "gui/comp.h"

namespace GUI
{
    /**
     * A point that moves over time.
     */
    class Movable_Point : public Updatable
    {
      public:
        /**
         * Construct a new point at (0, 0).  A new point will always
         * be stopped.
         */
        Movable_Point();
        
        /**
         * Move to the new location over time.  As the point moves,
         * the current location will change until it reaches its final
         * destination.  If <tt>milliseconds</tt> is 0, then the point
         * is moved immediate to the destination without actually
         * moving.
         *
         * @param x Destination x coordinate.
         * @param y Destination y coordinate.
         *
         * @parm milliseconds The time to move from the current
         * position to the destination, in milliseconds.
         */
        void move_to(int x, int y, int milliseconds);

        /**
         * Gets the current x coordinate.
         *
         * @return The current x coordinate.
         */
        int get_x() const;

        /**
         * Gets the current y coordinate.
         *
         * @return The current y coordinate.
         */
        int get_y() const;

        /**
         * Checks to see if the point is currently in motion.
         *
         * @return True if the point is stopped, false if it is moving.
         */
        bool is_stopped() const;
        
        /**
         * Update the current location of the point.
         *
         * @return True if the location has changed.
         */
        virtual bool update();

      private:
        /**
         * All possible states of a movable point.
         */
        enum State
        {
            /** This point is in motion to its destination. */
            MOVING,
            /** This point is not moving. */
            STOPPED
        };
        
        /** The current x coordinate. */
        fix p_x;

        /** The current y coordinate. */ 
        fix p_y;

        /**
         * The number of pixels per update to move this point in the x
         * direction.
         */
        fix p_velocity_x;

        /**
         * The nubmer of pixels per update to move this point in the y
         * direction.
         */
        fix p_velocity_y;

        /** The x destination while moving. */
        fix p_destination_x;

        /** The y destination when moving. */
        fix p_destination_y;

        /** The current state. */
        State p_state;
    };

    /**
     * A color that fades to and from black over time creating a
     * pulsating effect.
     */
    class Pulsating_Color : public Updatable
    {
      public:
        /**
         * Constructs a new pulsating color.  The default color is
         * white and it is constructed with a pulse period of 0
         * (i.e. not pulsating).
         */
        Pulsating_Color();

        /**
         * Sets the base color and pulsating time period.  The color
         * will constantly fade to black and then back again.
         * <tt>pulse_period</tt> is the time it takes to make one
         * cycle from color to black and back to color.  If it is 0,
         * then the color does not pulsate and always returns the base
         * color.
         *
         * @param color The base color.
         * @param pulse_period Time, in milliseconds, to fade to
         * black and back.
         */
        void set_color(int color, int pulse_period);

        /**
         * Gets the current color.
         *
         * @return The current color.
         */
        int get_color() const;

        /**
         * Changes the color, if needed.
         *
         * @return True if the color has changed from the last
         * update.
         */
        virtual bool update();
        
      private:
        /** The red component of the base color. */
        int p_base_r;

        /** The green component of the base color. */
        int p_base_g;

        /** The blue component of the base color. */
        int p_base_b;

        /** The red component of the current color. */
        fix p_current_r;

        /** The green component of the current color. */
        fix p_current_g;

        /** The blue component of the current color. */
        fix p_current_b;

        /** The amount the blue component changes per update. */
        fix p_delta_r;

        /** The amount the green component changes per update. */
        fix p_delta_g;

        /** The amount the blue component changes per update. */
        fix p_delta_b;

        /**
         * The direction the current color is move. If it is -1, then
         * the color is moving towards 0.  If it is 1, then it is move
         * towards the base color.
         */
        int p_direction;
    };

    /**
     * Represents the four points of a rectangle.
     */
    class Rectangle
    {
      public:
        /**
         * Constructs a rectangle with a upper left corner of (0, 0)
         * and a height and width of 1.
         */
        Rectangle();

        /**
         * Constructs a rectangle from the upper left corner, width,
         * and height.
         *
         * @param x The x-coordinate of the upper left corner.
         * @param y The y-coordinate of the upper left corner.
         * @param width The width.
         * @param height The height.
         */
        Rectangle(int x, int y, int width, int height);

        /**
         * Sets the upper left corner, width and height.
         */
        void set_bounds(int x, int y, int width, int height);

        /**
         * Gets the x-coordinate of the upper left corner.
         *
         * @return The x-coordinate of the upper left corner.
         */
        int get_x() const;

        /**
         * Gets the y-coordinate of the upper left corner.
         *
         * @return The y-coordinate of the upper left corener.
         */
        int get_y() const;

        /**
         * Gets the x-coordinate of the lower right corner.
         *
         * @return The x-coordinate of the lower right corner.
         */
        int get_lower_right_x() const;

        /**
         * Gets the y-coordinate of the lower right corner.
         *
         * @return The y-coordinate of the lower right corner.
         */
        int get_lower_right_y() const;

        /**
         * Gets the width.
         *
         * @return The width.
         */
        int get_width() const;

        /**
         * Gets the height.
         *
         * @r The height.
         */
        int get_height() const;

      private:
        /** The x-coordinate of the upper left corner. */
        int p_x;

        /** The y-coordinate of the upper left corner. */
        int p_y;

        /** The width. */
        int p_width;

        /** The height. */
        int p_height;
    };
};

#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
