#ifndef _MENU_H_
#define _MENU_H_

#include "allegro.h"

#ifdef __cplusplus
#include "engine.h"

#include <list>
#include <vector>

namespace GameLauncher
{
    using namespace GameEngine;

    class Screen_Shot;

    class Selection : public Drawable, public Updatable
    {
      public:
        enum State {SCROLLING, STOPPED};

        Selection(Engine * engine, FONT * new_font, BITMAP * icon);
        virtual ~Selection();

        virtual void    draw(BITMAP * frame);
        virtual bool    update();

        void            set_text(const string & text,
                                 bool show_icon = false);
        void            set_font(FONT * new_font);
        void            set_icon(BITMAP * icon);
        void            select();
        void            unselect();
        void            scroll_to(int x, int y, int milliseconds);
        void            move_to(int x, int y);
        int             x();
        int             y();
        int             s_x();
        int             s_y();
        bool            is_stopped();
        void            set_clip(int top, int bottom);
        
        static int      default_unselected_body_color;
        static int      default_unselected_outline_color;
        static int      default_selected_body_color;
        static int      default_selected_outline_color;
        static const int        OUTLINE_PADDING = 4;

      private:
        static const int    MAX_TEXT_LEN = 256;
        void            update_bitmap();

        Engine *        p_engine;
        FONT *          p_font;
        BITMAP *        p_icon;

        char            p_text[MAX_TEXT_LEN+1];
        int             p_pixel_width;
        int             p_pixel_height;
        bool            p_show_icon;

        int             p_unselected_body_color;
        int             p_unselected_outline_color;
        int             p_selected_body_color;
        int             p_selected_outline_color;

        bool            p_selected;
        
        bool            p_use_rle_sprite;
        BITMAP *        p_bitmap;
        RLE_SPRITE *    p_rle_sprite;
        fix             p_x;
        fix             p_y;
        fix             p_x_velocity;
        fix             p_y_velocity;
        fix             p_scroll_dest_x;
        int             p_scroll_dest_y;
        int             p_target_fps;
        fix             p_clip_top;
        fix             p_clip_bottom;

        State           p_state;
    };

    class Menu_Callback
    {
      public:
        virtual void    next_emu(unsigned int first_shown_index,
                                 unsigned int selected_index) = 0;
        virtual void    previous_emu(unsigned int first_shown_index,
                                     unsigned int selected_index) = 0;
        virtual void    select(unsigned int first_shown_index,
                               unsigned int selected_index) = 0;
        virtual void    quit(unsigned int first_shown_index,
                             unsigned int selected_index) = 0;
        virtual void    update_screen_shot(unsigned int selected_index,
                                           int brightness) = 0;
        virtual bool    screen_shot_available(unsigned int selected_index) = 0;
    };

    class Menu : public Drawable, public Updatable
    {
      public:
        Menu(Engine * engine, Screen_Shot * Screen_Shot, FONT * new_font,
             BITMAP * screen_shot_icon);
        virtual ~Menu();
    
        virtual void    draw(BITMAP * frame);
        virtual bool    update();

        SAMPLE *        next_emu_sample();
        void            next_emu_sample(SAMPLE * new_sample);
        SAMPLE *        previous_emu_sample();
        void            previous_emu_sample(SAMPLE * new_sample);
        SAMPLE *        selection_moved_sample();
        void            selection_moved_sample(SAMPLE * new_sample);
        SAMPLE *        game_chosen_sample();
        void            game_chosen_sample(SAMPLE * new_sample);
        SAMPLE *        quit_sample();
        void            quit_sample(SAMPLE * new_sample);
        void            screen_shot_delay(int delay);
        void            screen_shot_brightness(int brightness_percent);

        void            set_menu(const string & title,
                                 const string & sub_title,
                                 const vector<string> & all_selections,
                                 Menu_Callback * callback,
                                 unsigned int first_shown_index,
                                 unsigned int selected_index);

        void            set_key_repeat_delay(int initial_delay,
                                             int repeat_delay);
        void            set_up_key(int scancode);
        void            set_down_key(int scancode);
        void            set_page_up_key(int scancode);
        void            set_page_down_key(int scancode);
        void            set_play_game_key(int scancode);
        void            set_next_emu_key(int scancode);
        void            set_previous_emu_key(int scancode);
        void            set_quit_key(int scancode);
        void            set_screen_shot_key(int scancode);

      private:
        enum State {INIT,
                    INIT_WAIT,
                    SCROLLING_ON,
                    SCROLLING_UP_OR_DOWN,
                    TAKING_ORDERS,
                    DISPLAYING_SCREEN_SHOT,
                    SCROLLING_OFF_RIGHT,
                    SCROLLING_OFF_LEFT,
                    DONE_WAIT,
                    DONE,
                    STOPPED};

        typedef list<Selection *> Selection_List;

        static void     install_key_buffer();
        static void     kb_handler(int scancode);
        static bool     key_buffer_empty();
        static int      next_key_in_buffer();
        static void     clear_key_buffer();

        static const int    p_lr_scroll_ms  = 400;
        static const int    p_vert_scroll_ms = 100;
        static const int    p_separator_height = 8;

        bool            take_orders();
        void            set_selection(int num, char * selection);
        bool            are_all_selections_stopped();
        void            move_selection_up(unsigned int shift,
                                          unsigned int scroll);
        void            move_selection_down(unsigned int shift,
                                            unsigned int scroll);
        void            free_key(Keypress_Watcher * watcher);
        Keypress_Watcher *      get_repeat_key(int scancode);
        Keypress_Watcher *      get_one_time_key(int scancode);
        Joystick_Watcher *      get_one_time_joy(int scancode);

        Engine *        p_engine;
        FONT *          p_font;
        int             p_font_height;
        SAMPLE *        p_next_emu;
        SAMPLE *        p_previous_emu;
        SAMPLE *        p_selection_moved;
        SAMPLE *        p_game_chosen;
        SAMPLE *        p_quit;
        BITMAP *        p_screen_shot_icon;

        int             p_screen_shot_delay;
        int             p_screen_shot_brightness;

        Selection *     p_title;
        Selection *     p_sub_title;
                        
        Selection_List  p_selections;
        vector<string>  p_all_selections;
        Selection_List::iterator p_selected;
        unsigned int    p_selected_index;
        unsigned int    p_first_shown_index;
        unsigned int    p_num_shown;
        Selection_List  p_scroll_reserves;
                        
        Menu_Callback * p_menu_callback;
                        
        Timer *         p_init_timer;
        Timer *         p_done_timer;
        Timer *         p_screen_shot_timer;
        Screen_Shot *   p_screen_shot;
        State           p_state;
        int             p_done_delay;
                        
        int             p_y_top;
        int             p_y_bottom;

        Keypress_Watcher *      p_down_key;
        Keypress_Watcher *      p_up_key;
        Keypress_Watcher *      p_page_up_key;
        Keypress_Watcher *      p_page_down_key;
        Keypress_Watcher *      p_play_game_key;
        Keypress_Watcher *      p_next_emu_key;
        Keypress_Watcher *      p_previous_emu_key;
        Keypress_Watcher *      p_quit_key;
        Keypress_Watcher *      p_screen_shot_key;
        int             p_key_initial_delay;
        int             p_key_repeat_delay;
    };
}


extern "C" {
#endif /* __cplusplus */

void outline_textout(BITMAP * bmp, FONT * f, char * s,
                     int x, int y, int main_color, int outline_color);

#ifdef __cplusplus
};
#endif

#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
