#ifndef _VIEW_H_
#define _VIEW_H_

#include <vector>
#include <list>
#include "engine.h"

namespace GameLauncher
{
    using namespace GameEngine;

    class Movable_Bitmap;
    class Text_VC;
    class HText_VC;
    class VText_VC;
    class Scrollable_List;
    class Scrollable_List_Listener;
    class Null_Scrollable_List_Listener;
    class Delay_VC;
    class Delay_VC_Listener;

    class Movable_Bitmap : public Drawable, public Updatable
    {
      public:
        Movable_Bitmap(Engine * engine, int width, int height);
        virtual ~Movable_Bitmap();

        virtual void    draw(BITMAP * frame);
        virtual bool    update();

        void            move_to(int x, int y, int milliseconds = 0);
        bool            is_stopped();
        void            set_clip(int left, int top, int right, int bottom);
        BITMAP *        get_bitmap();
        
      protected:
        enum State {MOVING, STOPPED};

        Engine *        p_engine;

        BITMAP *        p_bitmap;
        int             p_width;
        int             p_height;
        fix             p_x;
        fix             p_y;
        fix             p_x_velocity;
        fix             p_y_velocity;
        fix             p_move_dest_x;
        int             p_move_dest_y;
        fix             p_clip_left;
        fix             p_clip_top;
        fix             p_clip_right;
        fix             p_clip_bottom;

        State           p_state;
    };

    class Text_VC : public Movable_Bitmap
    {
      public:
        Text_VC(Engine * engine, int width, int height, FONT * font);
        virtual         ~Text_VC() { }

        void            set_colors(int body_color, int outline_color);
        void            set_contents(const string & text, BITMAP * icon = 0);

        static const int        OUTLINE_PADDING = 4;

      protected:
        static const int        MAX_TEXT_LEN = 256;

        static bool     s_initialized_palette;
        static RGB_MAP  s_rgb_table;

        BITMAP *        get_outline_bmp();
        virtual void    update_bitmap() = 0;

        char            p_text[MAX_TEXT_LEN + 1];
        BITMAP *        p_icon;
        FONT *          p_font;
        int             p_body_color;
        int             p_outline_color;
        int             p_text_width;
    };

    class HText_VC : public Text_VC
    {
      public:
        HText_VC(Engine * engine, int width, int height, FONT * font);

      protected:
        void            update_bitmap();
    };

    class VText_VC : public Text_VC
    {
      public:
        VText_VC(Engine * engine, int width, int height, FONT * font);

      protected:
        void            update_bitmap();
    };

    class Outline_Bar : public Drawable
    {
      public:
        Outline_Bar(int x, int y, int width, int height);

        void    draw(BITMAP * bitmap);
        void    set_color(int color);

        static const int        OUTLINE_WIDTH = 2;

      protected:
        int     p_color;
        int     p_x1;
        int     p_y1;
        int     p_x2;
        int     p_y2;
        int     p_inner_x1;
        int     p_inner_y1;
        int     p_inner_x2;
        int     p_inner_y2;
    };

    class Scrollable_Item
    {
      public:
        string          text;
        BITMAP *        icon;
    };

    typedef vector<Scrollable_Item>     Scrollable_Item_List;

        /**
         * Possible axis values.
         */
    enum Mouse_Input_Axis {NO_AXIS,
                           POSITIVE_X_AXIS,
                           NEGATIVE_X_AXIS,
                           POSITIVE_Y_AXIS,
                           NEGATIVE_Y_AXIS};
        
    class Scrollable_List_VC : public Updatable, public Drawable
    {
      public:
        virtual ~Scrollable_List_VC();

        void            draw(BITMAP * frame);
        bool            update();

        void            set_selected_colors(int body, int outline);
        void            set_unselected_colors(int body, int outline);

        void            set_all_items(const Scrollable_Item_List & items,
                                      unsigned selected_index,
                                      unsigned first_shown_index);
        void            set_item_at(const Scrollable_Item & item,
                                    unsigned index);
        unsigned        get_selected_index();
        unsigned        get_first_shown_index();

        const Scrollable_Item * get_item_at(unsigned index);

        void            set_listener(Scrollable_List_Listener * listener);
        void	        set_select_joy(int joystick_event);
        void            set_line_up_joy(int joystick_event);
        void            set_line_down_joy(int joystick_event);
        void            set_page_up_joy(int joystick_event);
        void            set_page_down_joy(int joystick_event);
             
        void            set_line_up_key(int scancode);
        void            set_line_down_key(int scancode);
        void            set_page_up_key(int scancode);
        void            set_page_down_key(int scancode);
        void            set_select_key(int scancode);
   
        void            set_key_repeat_delay(int initial, int repeat);
        void            set_joystick_repeat_delay(int initial, int repeat);
        void            set_use_animations(bool use_animations);
        void            set_turbo_scroll_threshold(int threshold);

        void            hide_left(int milliseconds);
        void            hide_right(int milliseconds);
        void            hide_split(int milliseconds);
        void            center_selected_hide_others(int milliseconds);
        void            unhide_all(int milliseconds);
        bool            are_all_items_stopped();
        
        /**
         * Sets the mouse axis to be used as an input device.  If set
         * to NO_AXIS, then the mouse will not be used as an input
         * device.
         *
         * @param axis Input axis.
         */
        void set_mouse_input_axis(Mouse_Input_Axis axis);

        /**
         * Sets the threshold, in pixels, the mouse must move in order
         * to be registered as movement.
         *
         * @param threshold New threshold.
         */
        void set_mouse_input_threshold(int threshold);

      protected:
        enum State {HIDDEN,
                    UNHIDDEN,
                    SCROLLING,
                    GOING_INTO_HIDING,
                    COMING_OUT_OF_HIDING};

        enum Input_Command {NOTHING,
                            LINE_UP,
                            LINE_DOWN,
                            PAGE_UP,
                            PAGE_DOWN,
                            MOUSE_MOVE_UP,
                            MOUSE_MOVE_DOWN};

        typedef list<Text_VC *> Text_VC_List;

        static const int        DRAW_PRIORITY           = 5;
        static const int        UPDATE_PRIORITY         = 5;
        static const int        SCROLL_DELAY            = 100;
        static const int        DEFAULT_TURBO_THRESHOLD = 3;

        Scrollable_List_VC();

        virtual void    create_view_components() = 0;
        void            get_index_coordinates(int index, int * x, int * y);

        void            initialize(Engine * engine, FONT * font, int x, int y,
                                   int width, int height);
        Keypress_Watcher *      get_repeat_key(int scancode);
        Keypress_Watcher *      get_one_time_key(int scancode);
        Joystick_Watcher *      get_repeat_joy(int joystick_event);
        Joystick_Watcher *      get_one_time_joy(int joystick_event);
        Joystick_Event *        get_joystick_event(int joystick_event);
        void            initialize_container();
        bool            handle_input();
        void            update_visible_items();
        void            update_offscreen_items();
        void            move_selection_up(unsigned move, unsigned scroll);
        void            move_selection_down(unsigned move, unsigned scroll);
        void            hide_offscreen_items();
        bool            disable_scroll_delay();
        int             get_mouse_input_quantum();

        Engine *        p_engine;
        FONT *          p_font;
        GEContainer *   p_container;

        int             p_x;
        int             p_y;
        int             p_width;
        int             p_height;
        int             p_item_height;

        int             p_selected_body_color;
        int             p_selected_outline_color;
        int             p_unselected_body_color;
        int             p_unselected_outline_color;

        Scrollable_Item_List    p_items;
        unsigned        p_selected_index;
        unsigned        p_first_shown_index;

        Keypress_Watcher *      p_line_up_key;
        Keypress_Watcher *      p_line_down_key;
        Keypress_Watcher *      p_page_up_key;
        Keypress_Watcher *      p_page_down_key;
        Keypress_Watcher *      p_select_key;
        Mouse_Input_Axis        p_mouse_input_axis;
        int             p_contiguous_page_up_count;
        int             p_contiguous_page_down_count;
        int             p_turbo_scroll_threshold;
        int             p_mouse_input_threshold;
        Input_Command   p_last_input_command;

    
        bool                    p_use_joystick;
        Joystick_Watcher *      p_select_joy;    
        Joystick_Watcher *      p_line_up_joy;
        Joystick_Watcher *      p_line_down_joy;
        Joystick_Watcher *      p_page_up_joy;
        Joystick_Watcher *      p_page_down_joy;
    
        int             p_initial_joy_delay;
        int             p_repeat_joy_delay;
        
        int             p_initial_key_delay;
        int             p_repeat_key_delay;
        int             p_scroll_delay;

        Scrollable_List_Listener *      p_listener;
        Scrollable_List_Listener *      p_null_listener;

        Text_VC_List    p_visible_items;
        Text_VC_List    p_offscreen_items;
        int             p_num_visible_items;
        Text_VC_List::iterator  p_selected_item;
        State           p_state;

        int             p_visible_top_x;
        int             p_visible_top_y;
        int             p_visible_delta_x;
        int             p_visible_delta_y;
        int             p_visible_hidden_left_x;
        int             p_visible_hidden_left_y;
        int             p_visible_hidden_right_x;
        int             p_visible_hidden_right_y;
        int             p_offscreen_hidden_x;
        int             p_offscreen_hidden_y;
        int             p_centered_x;
        int             p_centered_y;
    };

    class Scrollable_List_Listener
    {
      public:
        virtual         ~Scrollable_List_Listener() { }
        virtual void    selection_changed(unsigned index) = 0;
        virtual void    item_selected(unsigned index) = 0;
        virtual void    item_visible(unsigned index) = 0;
    };

    class Null_Scrollable_List_Listener : public Scrollable_List_Listener
    {
      public:
        void            selection_changed(unsigned index);
        void            item_selected(unsigned index);
        void            item_visible(unsigned index);
    };

    class Scrollable_HList_VC : public Scrollable_List_VC
    {
      public:
        Scrollable_HList_VC(Engine * engine, FONT * font, int x, int y,
                            int width, int height);

      protected:
        void            create_view_components();
    };

    class Scrollable_VLList_VC : public Scrollable_List_VC
    {
      public:
        Scrollable_VLList_VC(Engine * engine, FONT * font, int x, int y,
                             int width, int height);

      protected:
        void            create_view_components();
    };

    class Menu_Listener
    {
      public:
        virtual         ~Menu_Listener() { }
        virtual void    selection_changed(unsigned index) = 0;
        virtual void    item_selected(unsigned index) = 0;
        virtual void    item_visible(unsigned index) = 0;
        virtual void    quit_pressed() = 0;
        virtual void    next_menu_pressed() = 0;
        virtual void    previous_menu_pressed() = 0;
        virtual void    stopped_moving() = 0;
    };

    class Null_Menu_Listener : public Menu_Listener
    {
      public:
        void            selection_changed(unsigned index);
        void            item_selected(unsigned index);
        void            item_visible(unsigned index);
        void            quit_pressed();
        void            next_menu_pressed();
        void            previous_menu_pressed();
        void            stopped_moving();
    };

    class Menu_View : public Drawable, public Updatable, 
        public Scrollable_List_Listener
    {
      public:
        virtual         ~Menu_View();
        void            set_titles(const string & title,
                                   const string & sub_title);

        void            set_selected_colors(int body, int outline);
        void            set_unselected_colors(int body, int outline);

        void            set_all_items(const Scrollable_Item_List & items,
                                      unsigned selected_index,
                                      unsigned first_shown_index);
        void            set_item_at(const Scrollable_Item & item,
                                    unsigned index);
        unsigned        get_selected_index();
        unsigned        get_first_shown_index();

        const Scrollable_Item * get_item_at(unsigned index);

        void            set_select_joy(int joystick_event);
        void            set_line_up_joy(int joystick_event);
        void            set_line_down_joy(int joystick_event);
        void            set_page_up_joy(int joystick_event);
        void            set_page_down_joy(int joystick_event);
        void            set_joystick_repeat_delay(int initial, int repeat);
       
        void            set_line_up_key(int scancode);
        void            set_line_down_key(int scancode);
        void            set_page_up_key(int scancode);
        void            set_page_down_key(int scancode);
        void            set_select_key(int scancode);
        void            set_quit_key(int scancode);
        void            set_next_menu_key(int scancode);
        void            set_previous_menu_key(int scancode);
        void            set_key_repeat_delay(int initial, int repeat);
        void            set_use_animations(bool use_animations);
        void            set_turbo_scroll_threshold(int threshold);
        void            set_listener(Menu_Listener * listener);
        void            hide_left(int milliseconds);
        void            hide_right(int milliseconds);
        void            hide_split(int milliseconds);
        void            center_selected_hide_others(int milliseconds);
        void            unhide_all(int milliseconds);

        // Scrollable_List_Listener interface
        void            selection_changed(unsigned index);
        void            item_selected(unsigned index);
        void            item_visible(unsigned index);

        void            draw(BITMAP * frame);
        bool            update();

        /**
         * Sets the mouse axis to be used as an input device.  If set
         * to NO_AXIS, then the mouse will not be used as an input
         * device.
         *
         * @param axis Input axis.
         */
        void set_mouse_input_axis(Mouse_Input_Axis axis);

        /**
         * Sets the threshold, in pixels, the mouse must move in order
         * to be registered as movement.
         *
         * @param threshold New threshold.
         */
        void set_mouse_input_threshold(int threshold);

      protected:
        enum State {HIDDEN,
                    UNHIDDEN,
                    SCROLLING,
                    GOING_INTO_HIDING,
                    COMING_OUT_OF_HIDING};

        static const int        SEPARATOR_HEIGHT        = 6;
        static const int        DRAW_PRIORITY           = 5;
        static const int        UPDATE_PRIORITY         = 5;

        Menu_View();
        void            initialize(Engine * engine, FONT * font, int x, int y,
                                   int width, int height);

        virtual void    create_view_components() = 0;
        void            hide_titles(int milliseconds);
        void            unhide_titles(int milliseconds);

        Keypress_Watcher *      get_one_time_key(int scancode);
        void            handle_keypresses();

        Engine *        p_engine;
        FONT *          p_font;
        
        int             p_x;
        int             p_y;
        int             p_width;
        int             p_height;
        int             p_item_height;

        GEContainer *   p_container;
        Text_VC *       p_title;
        Text_VC *       p_sub_title;
        Outline_Bar *   p_separator;
        Scrollable_List_VC *    p_list;

        Keypress_Watcher *      p_quit_key;
        Keypress_Watcher *      p_next_menu_key;
        Keypress_Watcher *      p_previous_menu_key;

        Menu_Listener * p_listener;
        Menu_Listener * p_null_listener;

        State           p_state;

        int             p_title_x;
        int             p_title_y;
        int             p_sub_title_x;
        int             p_sub_title_y;
        int             p_separator_x;
        int             p_separator_y;
        int             p_title_hidden_x;
        int             p_title_hidden_y;
        int             p_sub_title_hidden_x;
        int             p_sub_title_hidden_y;
    };

    class HMenu_View : public Menu_View
    {
      public:
        HMenu_View(Engine * engine, FONT * font, int x, int y, int width,
                   int height);

      protected:
        void            create_view_components();
    };

    class VLMenu_View : public Menu_View
    {
      public:
        VLMenu_View(Engine * engine, FONT * font, int x, int y, int width,
                   int height);

      protected:
        void            create_view_components();
    };


    class Delay_VC : public Updatable
    {
      public:
        Delay_VC(Engine * engine, Delay_VC_Listener * listener);
        virtual ~Delay_VC();
        void            delay(int milliseconds);

        // Updatable interface
        bool            update();

      private:
        Timer *         p_delay_timer;
        Delay_VC_Listener *     p_listener;
    };


    class Delay_VC_Listener
    {
      public:
        virtual         ~Delay_VC_Listener() { }
        virtual void    delay_vc_expired() = 0;
    };

    class Splash_VC : public Drawable, public Updatable,
                      public Delay_VC_Listener
    {
      public:
        Splash_VC(Engine *, Delay_VC_Listener * listener, FONT * font,
                  const string & text);
        virtual ~Splash_VC();

        void            splash(int milliseconds);

        bool            update();
        void            draw(BITMAP * frame);
        void            delay_vc_expired();


      private:
        Engine *        p_engine;
        Delay_VC *      p_delay_vc;
        Delay_VC_Listener *     p_listener;
        FONT *          p_font;
        string          p_text;
    };

    class Background_Music_Listener;
    class Null_Background_Music_Listener;

    /**
     * View component for the background music.
     */
    class Background_Music_VC : public Updatable
    {
      public:
        Background_Music_VC(Engine * engine);
        ~Background_Music_VC();

        virtual bool update();

        /**
         * Sets the scancode for the key to skip to the next song.
         *
         * @param scancode Key scancode.
         */
        void set_skip_key(int scancode);

        /**
         * Sets the scancode for the key to start the current song
         * over.
         *
         * @param scancode Key scancode.
         */
        void set_start_over_key(int scancode);

        /**
         * Sets the current background music listener.
         *
         * @param listener New listener, or 0 if no listener.
         */
        void set_listener(Background_Music_Listener * listener);

      private:
        /** Watches for skip key. */
        Keypress_Watcher *  p_skip_key;

        /** Watches for start over key. */
        Keypress_Watcher *  p_start_over_key;

        /** Current listener. */
        Background_Music_Listener * p_listener;

        /** Null listener. */
        Null_Background_Music_Listener * p_null_listener;
    };

    /**
     * An event for background music.
     */
    class Background_Music_Event
    {
    };

    /**
     * An interface for listeners of background music events.
     */
    class Background_Music_Listener
    {
      public:
        /**
         * Virtual destructor.
         */
        virtual ~Background_Music_Listener();

        /**
         * The skip song key was pressed.
         *
         * @param e Event.
         */
        virtual void skip_to_next_song(const Background_Music_Event * e) = 0;

        /**
         * The start song over key was pressed.
         * @param e Event.
         */
        virtual void start_song_over(const Background_Music_Event * e) = 0;
    };

    /**
     * An empty background music listener.
     */
    class Null_Background_Music_Listener : public Background_Music_Listener
    {
        virtual void skip_to_next_song(const Background_Music_Event * e);
        virtual void start_song_over(const Background_Music_Event * e);
    };

    class Screen_Saver_Listener;

    /**
     * View component for the screen saver.  This component controls
     * when the screen saver starts and stops as well as draws the
     * actual screen saver when it is active.
     */
    class Screen_Saver_VC : public Updatable, public Drawable
    {
      public:
        /**
         * Create a new screen saver view controller.
         *
         * @param engine Game engine
         * @param font Font to use while drawing.
         */
        Screen_Saver_VC(Engine * engine, FONT * font);

        /**
         * Destructor.
         */
        ~Screen_Saver_VC();

        /**
         * Checks the idel timer to see when the screen saver should
         * start and stop.
         */
        virtual bool update();

        /**
         * Draws the screen saver to the screen, if active.
         *
         * @param frame Screen frame.
         */
        virtual void draw(BITMAP * frame);

        /**
         * Sets the current screen saver event listener.
         *
         * @param listener New listener.
         */
        void set_screen_saver_listener(Screen_Saver_Listener * listener);

        /**
         * Sets the number of seconds before the screen saver becomes
         * active.  Setting the number of seconds to 0 disables the
         * screen saver.
         *
         * @param seconds Idle seconds before activating.
         */
        void set_start_after(unsigned int seconds);

        /**
         * Sets the text to display while the screen saver is active.
         * The default is the empty string.
         *
         * @param text Screen saver text
         */
        void set_text(string text);

      private:
        /**
         * Sends a changed state event to the listener.
         */
        void fire_changed_state_event(bool is_active);

        /**
         * Sets the text location to a random value.
         */
        void set_random_text_location();

        /** Reference to game engine. */
        Engine * p_engine;

        /** True if the screen saver is active. */
        bool p_is_active;

        /**
         * Number of seconds of idle time before screen saver becomes
         * active.
         */
        unsigned int p_start_after;
        
        /** The current listener. */
        Screen_Saver_Listener * p_listener;

        /** Font to use while drawing. */
        FONT * p_font;

        /** Text to be displayed. */
        string p_text;
        
        /** Timer that triggers random text relocation. */
        Timer * p_text_location_timer;

        /** X location of text. */
        int p_text_x;

        /** Y location text. */
        int p_text_y;
    };

    /**
     * An event for screen saver changes.
     */
    class Screen_Saver_Event
    {
      public:
        /**
         * Create a new event.
         *
         * @param is_active Active flag.
         */
        Screen_Saver_Event(bool is_active);

        /**
         * Gets the active flag.  If <tt>true</tt> then the screen
         * saver just became active.  If <tt>false</tt> then the
         * screen saver just became inactive.
         */
        bool is_active() const;

      private:
        /** The active flag. */
        bool            p_is_active;
    };

    /**
     * The interface that objects interested in screen saver events
     * must implement.
     */
    class Screen_Saver_Listener
    {
      public:
        /**
         * Destructor.
         */
        virtual ~Screen_Saver_Listener();

        /**
         * Called when the screen saver changes state.
         */
        virtual void screen_saver_changed_state(
            const Screen_Saver_Event * e) = 0;
    };

#if 0
    class Test_TVC : public Updatable
    {
      public:
        Test_TVC(Engine * engine, FONT * font);
        bool    update();

      private:
        Text_VC * p_tvc;
        FONT * p_font;
        Engine * p_engine;
        BITMAP * p_icon;
    };

    class Test_SL : public Updatable
    {
      public:
        Test_SL(Engine * engine, FONT * font);
        virtual ~Test_SL();
        bool    update();

      private:
        static const int        NUM_ITEMS = 45;
        void    set_item(unsigned item);

        Keypress_Watcher *      ks;
        Keypress_Watcher *      kd;
        Keypress_Watcher *      kf;
        Keypress_Watcher *      ke;
        Engine *        p_engine;
        FONT *          p_font;
        Menu_View *     p_menu;
        Scrollable_Item_List    p_items;
    };
#endif
}

#endif  /* _VIEW_H_ */

/* Local Variables: */
/* mode: c++ */
/* End: */
