#include <stdio.h>
#include <dirent.h>

#include "mamescan.h"
#include "options.h"

using namespace MameScan;


ConfigGenerator::ConfigGenerator(const string & infoFileName,
                                 const string & configFileBaseName,
                                 const string & romDirectory)
    : mGameList(), mMatchedGames(), mConstraints()
{
    mInfoFileName       = infoFileName;
    mConfigFileBaseName = configFileBaseName;
    mRomDirectory       = romDirectory;
}

ConfigGenerator::~ConfigGenerator()
{
    vector<Constraint *>::iterator  i;

    for (i = mConstraints.begin(); i != mConstraints.end(); i++)
    {
        Constraint  * constraint;

        constraint  = (*i);
        delete constraint;
    }
}

void
ConfigGenerator::generateConfig() throw (Exception)
{
    Options *   options;

    options = Options::instance();
    
    mGameList.parseInfoFile(mInfoFileName);
    parseConstraintFile(options->config_file);
    findGames();
    outputConfigFiles();
}

void
ConfigGenerator::parseConstraintFile(const string & constraintFileName)
{
    FILE *  file;

    file    = fopen(constraintFileName.c_str(), "r");

    if (file == NULL)
    {
        fprintf(stderr, "Cannot open constraint file, ignoring\n");
        return;
    }

    while (1)
    {
        char    line[1000];
        char    value[10];
        char    type;
        char *  first;
        char *  last;
        int     num_chars;

        if (fgets(line, sizeof(line), file) == NULL)
            break;

        first   = &line[0];

        while (isspace(*first))
            first++;
        
        // Type is the first non-whitespace character
        type    = *first;

        // Value starts just after type character, and goes until next
        // whitespace.
        first++;

        last = first;

        while (! isspace(*last))
            last++;
        
        num_chars   = MIN(size_t(last - first), sizeof(value) - 1);
        strncpy(value, first, num_chars);
        value[num_chars] = '\0';

        if (type == '+')
        {
            mConstraints.push_back(new IncludeConstraint(string(value)));
        }
        else if (type == '-')
        {
            mConstraints.push_back(new ExcludeConstraint(string(value)));
        }
    }

    fclose(file);
    return;
}

void
ConfigGenerator::findGames() throw (Exception)
{
    DIR *       dir;
    char *      extension;
    string      romName;
    GameInfo *  info;
    Options *   options;
    bool        added;
    Constraint::Result  constraint;

    struct dirent *     entry;
    vector<GameInfo *>  infoList;
    vector<GameInfo *>::iterator    i;

    options = Options::instance();
    dir = opendir(mRomDirectory.c_str());
    if (dir == NULL)
    {
        throw Exception("Unable to open " + mRomDirectory);
    }

    while ((entry = readdir(dir)))
    {
        splitExtension(entry->d_name, &extension);

        if ((extension != 0) && (strcmp(extension, "zip") == 0))
        {
            romName     = string(entry->d_name);
            if (options->verbose)
                printf("%s:\n", romName.c_str());
            info        = mGameList.lookupByName(romName);

            if (info == 0)
            {
                if (options->verbose)
                    printf("No info found\n\n");
                continue;
            }
            
            added   = false;

            // Always add a ROM with a direct match, assuming it is
            // good, even if it is a clone.  A constraint overrides
            // this check.

            constraint  = matchConstraints(info->name);

            if (constraint != Constraint::NO_CONSTRAINT)
            {
                if (constraint ==  Constraint::INCLUDE)
                {
                    mMatchedGames.push_back(info);
                    added= true;
                }

                // An EXCLUDE constraint does nothing, which keeps
                // this game out of the list.
            }
            else if (info->status == GOOD)
            {
                mMatchedGames.push_back(info);
                added = true;
            }

            if (options->verbose)
            {
                printf("%s", (added ? "[ADD]  " : "[SKIP] "));
                printGameInfo(info);
            }

            // Now check other games that may be buried within this
            // ROM.

            infoList    = mGameList.lookupByRomOf(romName);
            for (i = infoList.begin(); i != infoList.end(); i++)
            {
                info    = *i;
                added   = false;

                // Skip the clones unless specifically overridden.  A
                // constraint overrides all options.

                constraint  = matchConstraints(info->name);
                if (constraint != Constraint::NO_CONSTRAINT)
                {
                    if (constraint ==  Constraint::INCLUDE)
                    {
                        mMatchedGames.push_back(info);
                        added= true;
                    }

                    // An EXCLUDE constraint does nothing, which keeps
                    // this game out of the list.
                }
                else if ((! info->isClone()) || (options->use_clones))
                {
                    if (info->status == GOOD)
                    {
                        mMatchedGames.push_back(info);
                        added = true;
                    }
                }

                if (options->verbose)
                {
                    printf("%s", (added ? "[ADD]  " : "[SKIP] "));
                    printGameInfo(info);
                }

            }
            if (options->verbose)
                printf("\n");
        }
    }

    closedir(dir);
    return;
}

void
ConfigGenerator::outputConfigFiles() throw (Exception)
{
    GameInfo *  info;
    string      romFileName;
    string      mapFileName;
    FILE *      romFile;
    FILE *      mapFile;
    vector<GameInfo *>::iterator    i;

    romFileName = mConfigFileBaseName + ".rom";
    mapFileName = mConfigFileBaseName + ".map";

    romFile = fopen(romFileName.c_str(), "w");
    if (romFile == NULL)
    {
        throw Exception("Cannot create " + romFileName);
    }

    mapFile = fopen(mapFileName.c_str(), "w");
    if (mapFile == NULL)
    {
        fclose(romFile);
        throw Exception("Cannot create " + mapFileName);
    }

    for (i = mMatchedGames.begin(); i != mMatchedGames.end(); i++)
    {
        info    = *i;
        fprintf(romFile, "%s\n", info->name);
        fprintf(mapFile, "%s \"%s\"\n", info->name, info->description);
    }

    fclose(romFile);
    fclose(mapFile);
}


void
ConfigGenerator::printGameInfo(GameInfo * info)
{
    const char *    cloneString;
    const char *    statusString;


    cloneString     = (info->isClone() ? "[CLONE] " : "");
    statusString    = ((info->status == GOOD) ? "[G]" : "[B]");

    printf("%-8s -> %s %s%s\n", info->name, info->description,
           cloneString, statusString);
}

Constraint::Result
ConfigGenerator::matchConstraints(const string & game)
{
    Constraint::Result  result;
    vector<Constraint *>::iterator  i;

    for (i = mConstraints.begin(); i != mConstraints.end(); i++)
    {
        result  = (*i)->matchConstraint(game);

        // Exit at first constraint matched
        if (result != Constraint::NO_CONSTRAINT)
            break;
    }

    return result;
}
