#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#ifdef TARGET_UNIX
  #include <libgen.h>
#endif
#ifdef TARGET_MINGW
  #include <libiberty.h>
#endif

#include "mamescan.h"
#include "options.h"

using namespace MameScan;


static int  parse_options(int argc, char * argv[]);
static void show_usage(int exit_status);

const char * MameScan::program_name = 0;


/************************************************************************/


inline static bool isSeparator(char ch)
{
    return ((ch == '/') || (ch == '\\'));
}

void    splitExtension(char * path, char ** extension)
{
    int         i;
    bool        found;
    bool        done;

    i           = strlen(path) - 1;
    found       = false;
    done        = false;

    while (! done)
    {
        char    ch;

        ch = path[i];
        if (ch == '.')
        {
            found = true;
            break;
        }

        // Not found, so just break
        if ((isSeparator(ch)) || (i == 0))
            done = true;

        i--;
    }
    
    if (found)
    {
        path[i]     = '\0';
        *extension  = &path[i+1];
    }
    else
    {
        *extension  = 0;
    }
}


/************************************************************************/


int
main(int argc, char *argv[])
{                                  
    Options *       options;
    int         numOptions;
    int         numArguments;
    int         rc;

    options = Options::instance();
    program_name    = basename(argv[0]);

    numOptions      = parse_options(argc, argv);
    numArguments    = argc - numOptions - 1;

    if (options->show_help)
        show_usage(0);

    if (options->show_version)
    {
        printf("mamescan version %s\n", GLAUNCH_VERSION);
        printf("Please report any bugs to <drib@enteract.com>.\n");
        exit(0);
    }
    

    if (numArguments != 0)
    {
        fprintf(stderr, "%s: Too many arguments given\n", program_name);
        show_usage(1);
    }

    rc = 0;
    try
    {
        Scanner *   scanner;

        scanner = new Scanner(options->config_file);
        scanner->scan();
        delete scanner;
    }
    catch (Exception e)
    {
        fprintf(stderr, "%s: %s\n", program_name, e.getError().c_str());
        rc = 1;
    }

    return rc;
}


static void
show_usage(int exit_status)
{
    if (exit_status != 0)
    {
        fprintf(stderr, "Try `%s --help' for more information.\n",
                program_name);
    }
    else
    {
        printf("Usage: %s [OPTIONS]...\n\n",
               program_name);
        printf("\
Scans a MAME ROM directory and generates Game Launcher config files.\n\
\n\
  -h, --help                Displays this help and exits.\n\
      --version             Displays version information and exits.\n\
  -v, --verbose             Turns on verbose output.\n\
  -c FILE, --config-file FILE \n\
                            Specifies a configuration file to use.\n\
                            The default is mamescan.cfg.\n\
\n\
Please report any bugs to <drib@enteract.com>.\n");
    }

    exit(exit_status);
}

static int
parse_options(int      argc,
              char *   argv[])
{
    bool        done;
    Options *   opt = Options::instance();
    int         longind;

    struct option const long_options [] =
    {
        // Name         Arguments?      Variable                Value
        {"help",        no_argument,    &opt->show_help,        opt_true},
        {"version",     no_argument,    &opt->show_version,     opt_true},
        {"verbose",     no_argument,    &opt->verbose,          opt_true},
        {"config-file", required_argument,
                                        0,                      0},
        {"use-clones",  no_argument,    &opt->use_clones,       opt_true},
        {"dump-info",   no_argument,    &opt->dump_info,        opt_true},
        {"debug",       no_argument,    &opt->debug,            opt_true},
        {0,             0,              0,                      0}
    };

    done = false;
    while(! done)
    {
        int     c;

        c = getopt_long(argc, argv, "hvc:", long_options, &longind);
        if (c == -1)
        {
            done = true;
            continue;
        }
        
        switch (c)
        {
            case 0:
                if (strcmp(long_options[longind].name, "config-file") == 0)
                {
                    opt->config_file    = string(optarg);
                }
                break;

            case 'c':
                opt->config_file    = string(optarg);
                break;
                
            case 'h':
                opt->show_help  = opt_true;
                break;

            case 'v':
                opt->verbose    = opt_true;
                break;

            default:
                show_usage(1);
                break;
        }
    }

    return (optind - 1);
}
