#include <stdio.h>
#include <dirent.h>
#include <fstream>

#include "fileutil.h"
#include "glaunch.h"
#include "logfile.h"

using namespace GameLauncher;
using LogFile::gl_log;
using namespace std;

bool
nocase_less :: operator()(const string & s1, const string & s2)
{
    string::const_iterator      p1;
    string::const_iterator      p2;

    p1 = s1.begin();
    p2 = s2.begin();

    while ((p1 != s1.end()) && (p2 != s2.end()))
    {
        if (toupper(*p1) != toupper(*p2))
            return (toupper(*p1) < toupper(*p2)) ? true : false;
        p1++;
        p2++;
    }

    if (s1.size() < s2.size())
        return true;
    else
        return false;

}

char * Emulator::CONFIG_OPTIONS_SECTION         = "Options";

char * Emulator::CONFIG_SECTION                 = "Emulator";
char * Emulator::CONFIG_NAME                    = "name";
char * Emulator::CONFIG_DESCRIPTION             = "description";
char * Emulator::CONFIG_DIRECTORY               = "directory";
char * Emulator::CONFIG_EXECUTABLE              = "executable";
char * Emulator::CONFIG_USE_ROM_FILE            = "use_rom_file";
char * Emulator::CONFIG_ROM_EXTENSION           = "rom_extension";
char * Emulator::CONFIG_ROM_DIRECTORY           = "rom_directory";
char * Emulator::CONFIG_USE_EXTENSION           = "use_extension";
char * Emulator::CONFIG_USE_ROM_DIRECTORY       = "use_rom_directory";
char * Emulator::CONFIG_GLOBAL_OPTIONS          = "global_options";
char * Emulator::CONFIG_OPTIONS_BEFORE_ROM      = "options_before_rom";
char * Emulator::CONFIG_MAP_FILE                = "map_file";
char * Emulator::CONFIG_SCREEN_SHOT_DIRECTORY   = "screen_shot_directory";
char * Emulator::CONFIG_PRE_COMMAND             = "pre_command";
char * Emulator::CONFIG_POST_COMMAND            = "post_command";

char * Emulator::STATE_SELECTED_INDEX           = "selected_index";
char * Emulator::STATE_FIRST_SHOWN_INDEX        = "first_shown_index";

Emulator::Emulator()
    : p_name_map(), p_games()
{
}

Emulator::Emulator(const string & config_name,
                   const string & path_prefix,
                   Emulator_Listener * listener = 0
                   ) throw (InvalidEmulatorConfig)
    : p_name_map(), p_games()
{
    string      config_file_path;
    p_path_prefix       = path_prefix;
    p_config_directory  = "config";
    p_config_name       = config_name;
    config_file_path    = p_config_directory + "/" + p_config_name + ".cfg";
    fix_file_name_case(config_file_path);
    p_config_file       = new Config_File(config_file_path);

    if (listener != 0)
        p_emulator_listener     = listener;
    else
        p_emulator_listener     = Null_Emulator_Listener::instance();

    load_config();

    // Do some rudimentary error checking and bail if certain options
    // aren't set.
    if ((p_name == "") ||
        (p_executable == ""))
    {
        throw InvalidEmulatorConfig();
    }

    load_name_map();

    if (p_use_rom_file == true)
        add_games_from_rom_file();
    else
        search_for_games();

    if (p_selected_index >= p_games.size())
        p_selected_index = 0;
    if (p_first_shown_index >= p_games.size())
        p_first_shown_index = 0;

    return;
}

Emulator::~Emulator()
{
    GameListImpl::iterator      i;

    delete p_config_file;
    if (p_rom_extension != 0)
        delete p_rom_extension;

    for (i = p_games.begin(); i != p_games.end(); i++)
    {
        delete i->second;
    }
}

unsigned
Emulator::get_selected_index() const
{
    return p_selected_index;
}

unsigned
Emulator::get_first_shown_index() const
{
    return p_first_shown_index;
}

void
Emulator::set_selected_index(unsigned int selected_index_)
{
    p_selected_index = selected_index_;
    return;
}

void
Emulator::set_first_shown_index(unsigned int first_shown_index_)
{
    p_first_shown_index = first_shown_index_;
    return;
}

void
Emulator::restore_state(Config_File * state_file)
{
    Config_Section *    section;

    section             = state_file->get_section(p_config_name);
    p_selected_index    = section->get_int(STATE_SELECTED_INDEX, 0);
    p_first_shown_index = section->get_int(STATE_FIRST_SHOWN_INDEX, 0);
    delete section;

    return;
}

void
Emulator::save_state(Config_File * state_file)
{
    Config_Section *    section;

    section     = state_file->get_section(p_config_name);
    section->set_int(STATE_SELECTED_INDEX, p_selected_index);
    section->set_int(STATE_FIRST_SHOWN_INDEX, p_first_shown_index);
    delete section;

    return;
}

string
Emulator::get_screen_shot_file_name(Game * game)
{
    string              screen_shot_no_ext;
    string              pcx_file;
    string              bmp_file;
    string              png_file;
    string              jpg_file;
    string              screen_shot_file;

    screen_shot_no_ext = (p_screen_shot_directory + "\\" +
                          game->get_rom_name());
    pcx_file = screen_shot_no_ext + ".pcx";
    bmp_file = screen_shot_no_ext + ".bmp";
    png_file = screen_shot_no_ext + ".png";
    jpg_file = screen_shot_no_ext + ".jpg";

    screen_shot_file = "";
    if (file_exists(pcx_file))
        screen_shot_file = pcx_file;
#if (USE_PNG == 1)
    else if (file_exists(png_file))
        screen_shot_file = png_file;
#endif
    else if (file_exists(bmp_file))
        screen_shot_file = bmp_file;
#if (USE_JPG == 1)
    else if (file_exists(jpg_file))
        screen_shot_file = jpg_file;
#endif

    return screen_shot_file;
}

string
Emulator::get_command_to_run(Game * game)
{
    string      command;
    string      rom_file;
    string      rom;
    string      options;
    string      game_options;
    
    Config_Section *    section;

    section             = p_config_file->get_section(CONFIG_OPTIONS_SECTION);
    game_options        = section->get_string(game->get_rom_name(), "");
    delete section;

    if ((p_use_extension) && (! p_use_rom_file))
        rom_file = game->get_rom_name() + "." + game->get_extension();
    else
        rom_file = game->get_rom_name();

    if (p_use_rom_directory)
        rom = p_rom_directory + rom_file;
    else
        rom = rom_file;

    quote_long_file_name(rom);

    options = p_global_options + " " + game_options;

    if (p_options_before_rom)
        command = p_executable + " " + options + " " + rom;
    else
        command = p_executable + " " + rom + " " + options;

    return command;
}

const string &
Emulator::config_directory() const
{
    return p_config_directory;
}

const string &
Emulator::config_name() const
{
    return p_config_name;
}

const string &
Emulator::name() const
{
    return p_name;
}

const string &
Emulator::description() const
{
    return p_description;
}

const string &
Emulator::directory() const
{
    return p_directory;
}

const string &
Emulator::executable() const
{
    return p_executable;
}

const string &
Emulator::rom_directory() const
{
    return p_rom_directory;
}

const string &
Emulator::global_options() const
{
    return p_global_options;
}

const string &
Emulator::screen_shot_directory() const
{
    return p_screen_shot_directory;
}

const string &
Emulator::map_file() const
{
    return p_map_file;
}

bool
Emulator::options_before_rom() const
{
    return p_options_before_rom;
}

bool
Emulator::use_rom_file() const
{
    return p_use_rom_file;
}

bool
Emulator::use_extension() const
{
    return p_use_extension;
}

bool
Emulator::use_rom_directory() const
{
    return p_use_rom_directory;
}

const string &
Emulator::get_pre_command() const
{
    return p_pre_command;
}

const string &
Emulator::get_post_command() const
{
    return p_post_command;
}

void
Emulator::load_config()
{
    // Setup some defaults
    p_name                      = "";
    p_description               = "";
    p_directory                 = ".";
    p_pre_command               = "";
    p_executable                = "";
    p_post_command              = "";
    p_rom_directory             = ".";
    p_rom_extension             = 0;
    p_global_options            = "";
    p_screen_shot_directory     = ".";
    p_map_file                  = "";
    p_options_before_rom        = false;
    p_use_rom_file              = false;
    p_use_extension             = false;
    p_use_rom_directory         = false;


    Config_Section *    section;

    section             = p_config_file->get_section(CONFIG_SECTION);
    
    p_name              = section->get_string(CONFIG_NAME, p_name);
    p_description       = section->get_string(CONFIG_DESCRIPTION,
                                              p_description);
    p_directory         = section->get_file_name(CONFIG_DIRECTORY,
                                                 p_directory);
    p_directory         = append_prefix(p_directory);
    p_executable        = section->get_file_name(CONFIG_EXECUTABLE,
                                                 p_executable);
    p_pre_command       = section->get_string(CONFIG_PRE_COMMAND,
                                              p_pre_command);
    p_post_command      = section->get_string(CONFIG_POST_COMMAND,
                                              p_post_command);
    p_use_rom_file      = section->get_bool(CONFIG_USE_ROM_FILE,
                                            p_use_rom_file);

    if (p_use_rom_file == false)
    {
        p_rom_extension = section->get_file_name_list(CONFIG_ROM_EXTENSION);
        p_rom_directory = section->get_file_name(CONFIG_ROM_DIRECTORY,
                                                 p_rom_directory);
        add_backslash(p_rom_directory);
        p_rom_directory = append_prefix(p_rom_directory);
        fix_filename_slashes(p_rom_directory);
    }

    p_use_extension     = section->get_bool(CONFIG_USE_EXTENSION,
                                            p_use_extension);
    p_use_rom_directory = section->get_bool(CONFIG_USE_ROM_DIRECTORY,
                                            p_use_rom_directory);
    p_global_options    = section->get_string(CONFIG_GLOBAL_OPTIONS,
                                              p_global_options);
    p_options_before_rom = section->get_bool(CONFIG_OPTIONS_BEFORE_ROM,
                                             p_options_before_rom);
    p_map_file          = section->get_file_name(CONFIG_MAP_FILE, p_map_file);
    p_screen_shot_directory =
        section->get_file_name(CONFIG_SCREEN_SHOT_DIRECTORY,
                               p_screen_shot_directory);
    p_screen_shot_directory = append_prefix(p_screen_shot_directory);

    delete section;

    p_emulator_listener->poll_emulator_listener();

    return;
}

void
Emulator::load_name_map()
{
    string      map_file_path;
    string      line;
    string      short_name;
    string      full_name;

    map_file_path = p_config_directory + "/" + p_map_file;
    ifstream map_stream(map_file_path.c_str());
    if (! map_stream)
        return;

    while (! map_stream.eof())
    {
        string::size_type len, s_start, s_end, s_len, l_start, l_end, l_len;

        getline(map_stream, line);

        len = line.size();
        // s_start will point to first non-space character
        s_start = 0;
        while ((s_start < len) && (isspace(line[s_start])))
            s_start++;
        if (s_start == len)
            continue;

        // s_end will point to character before next space
        s_end = s_start+1;
        while ((s_end < len) && (! isspace(line[s_end])))
            s_end++;
        if (s_end == len)
            continue;
        s_end--;
        s_len = s_end - s_start + 1;
        short_name = line.substr(s_start, s_len);

        // l_start will point to first char after first double-quote
        l_start = s_end+1;
        while ((l_start < len) && (line[l_start] != '"'))
            l_start++;
        if (l_start == len)
            continue;
        // Skip double-quote
        l_start++;

        // l_end will point to next double-quote
        l_end = l_start+1;
        while ((l_end < len) && (line[l_end] != '"'))
            l_end++;
        if (l_end == len)
            continue;
        // Skip double-quote
        l_end--;
        l_len = l_end - l_start + 1;
        full_name = line.substr(l_start, l_len);

        fix_file_name_case(short_name);
        p_name_map[short_name] = full_name;
        p_emulator_listener->poll_emulator_listener();
    }
    map_stream.close();
    
    return;
}


void
Emulator::add_game(const string & rom_base, const string & extension)
{
    Game *      game;
    string      full_name;
    NameMap::iterator name_lookup;

    name_lookup = p_name_map.find(rom_base);
    if(name_lookup == p_name_map.end())
        full_name = rom_base;
    else
        full_name = name_lookup->second;

    game = new Game(full_name, rom_base, extension);
    // Possible memory leak if full_name is already in the map?
    p_games[full_name] = game;

    p_emulator_listener->poll_emulator_listener();

    return;
}

void
Emulator::add_games_from_rom_file()
{
    string      rom_file;
    string      rom;

    rom_file = (p_config_directory + "/" +
                p_config_name + ".rom");
    ifstream rom_stream(rom_file.c_str());
    if (! rom_stream)
    {
        gl_log->error("Could not open ROM file: " + rom_file);
        return;
    }

    while (! rom_stream.eof())
    {
        rom_stream >> rom;
        fix_file_name_case(rom);
        add_game(rom, "");
    }

    return;
}

void
Emulator::search_for_games()
{
    struct dirent *     entry;
    DIR *       dir;
    string      rom_file;
    string      rom_base = "";
    string      extension = "";
    
    // Walk the ROM directory looking for ROM files
    dir = opendir(p_rom_directory.c_str());
    if (dir == 0)
        return;
    while((entry = readdir(dir)))
    {
        rom_file = string(entry->d_name);
        fix_file_name_case(rom_file);

        split_extension(rom_file, rom_base, extension);
                            
        // Not a ROM file
        if (! is_valid_extension(extension))
            continue;
        add_game(rom_base, extension);
    }
    closedir(dir);
    
    return;
}

string
Emulator::append_prefix(const string & path)
{
    if (is_absolute_path(path))
        return path;
    else
        return (p_path_prefix + path);
}

bool
Emulator::is_valid_extension(const string & extension)
{
    bool        is_valid;
    vector<string>::iterator    i;

    if (p_rom_extension == 0)
        return false;

    is_valid    = false;
    for (i = p_rom_extension->begin(); i != p_rom_extension->end(); i++)
    {
        if (*i == extension)
        {
            is_valid    = true;
            break;
        }
    }

    return is_valid;
}

int
Emulator::number_of_games() const
{
    return p_games.size();
}

void
Emulator::get_game_list(Game_List & game_list) const
{
    GameListImpl::const_iterator  i;
    int     count;

    // Assuming a zero-size vector is passed in, this will stretch the
    // vector out to the correct size.  If a vector is reused, space
    // is not reclaimed if the resulting list is smaller.
    game_list.resize(p_games.size());

    count = 0;
    for (i = p_games.begin(); i != p_games.end(); i++)
    {
        game_list[count] = i->second;
        count++;
    }

    return;
}

InvalidEmulator::InvalidEmulator(const string & config_name,
                                 Emulator_Listener * listener = 0)
    : Emulator()
{
    if (listener != 0)
        p_emulator_listener     = listener;
    else
        p_emulator_listener     = Null_Emulator_Listener::instance();


    p_name          = "Invalid Configuration";
    p_description   = "(" + config_name + ")";

    p_selected_index    = 0;
    p_first_shown_index = 0;
}

Emulator_Listener::~Emulator_Listener()
{
}

Emulator_Listener * Null_Emulator_Listener::p_instance = 0;

void
Null_Emulator_Listener::poll_emulator_listener()
{
}

Emulator_Listener *
Null_Emulator_Listener::instance()
{
    if (p_instance == 0)
        p_instance = new Null_Emulator_Listener();

    return p_instance;
}
