#include "logfile.h"
#include "fileutil.h"
#include "glaunch.h"

using namespace GameLauncher;
using namespace std;
using LogFile::gl_log;

Emulator_Manager::Emulator_Manager(Config_File * config_file,
                                   Config_File * state_file)
    : p_emulator_list(), p_emulator_loaded()
{
    p_current_emulator  = 0;
    p_emulator_names    = 0;
    p_listener          = Null_Emulator_Manager_Listener::instance();

    p_state_file        = state_file;

    Config_Section *    section;

    section             = config_file->get_section(CONFIG_SECTION);
    p_emulator_names    = section->get_string_list(CONFIG_EMULATORS);
    p_path_prefix       = section->get_string(CONFIG_PATH_PREFIX, "");
    delete section;

    unsigned            i;
    unsigned            names_size;

    names_size  = p_emulator_names->size();
    p_emulator_list.resize(names_size);
    p_emulator_loaded.resize(names_size);
    p_current_emulator  = 0;

    for (i = 0; i < names_size; i++)
    {
        // cout << i << " emu " << (*p_emulator_names)[i] << endl;
        // Load the emulators as they are needed.
        p_emulator_list[i]      = 0;
        p_emulator_loaded[i]    = false;
    }
}

Emulator_Manager::~Emulator_Manager()
{
    if (p_emulator_names != 0)
        delete p_emulator_names;
}

void
Emulator_Manager::restore_state()
{
    unsigned            current_emulator;
    Config_Section *    section;

    section             = p_state_file->get_section(STATE_SECTION);
    current_emulator    = (unsigned) section->get_int(STATE_CURRENT_EMU, 0);
    delete section;

    if (current_emulator < p_emulator_list.size())
        p_current_emulator = current_emulator;

    p_listener->emulator_manager_changed();
}

void
Emulator_Manager::save_state()
{
    Emulator_List::iterator     i;

    for (i = p_emulator_list.begin(); i != p_emulator_list.end(); i++)
    {
        if ((*i) != 0)
            (*i)->save_state(p_state_file);
    }

    Config_Section *    section;
    section = p_state_file->get_section(STATE_SECTION);
    section->set_int(STATE_CURRENT_EMU, p_current_emulator);
    delete section;
}

void
Emulator_Manager::set_listener(Emulator_Manager_Listener * listener)
{
    if (listener != 0)
        p_listener = listener;
    else
        p_listener = Null_Emulator_Manager_Listener::instance();

    // Update new listener
    p_listener->emulator_manager_changed();
}

unsigned
Emulator_Manager::get_number_of_emulators()
{
    return p_emulator_list.size();
}

void
Emulator_Manager::load_emulator_if_necessary()
{
    if (! p_emulator_loaded[ p_current_emulator ])
    {
        Emulator *  emulator;
        string      name;

        name        = (*p_emulator_names)[ p_current_emulator ];
        // cout << "Loading " << name << endl;

        try
        {
            emulator = new Emulator(name, p_path_prefix, this);
            emulator->restore_state(p_state_file);
        }
        catch (InvalidEmulatorConfig e)
        {
            // Use an invalid emulator as a place holder
            gl_log->info("Invalid emulator config for " + name);
            emulator = new InvalidEmulator(name, this);
        }

        p_emulator_loaded[ p_current_emulator ] = true;
        p_emulator_list[ p_current_emulator ]   = emulator;
    }
}

Emulator *
Emulator_Manager::get_current_emulator()
{
    // cout << "Current " << p_current_emulator << endl;
    // cout << "Size " <<p_emulator_list.size() << endl;
    if (p_current_emulator < p_emulator_list.size())
    {
        load_emulator_if_necessary();
        return p_emulator_list[ p_current_emulator ];
    }
    else
        return 0;
}

unsigned
Emulator_Manager::get_current_emulator_index()
{
    return p_current_emulator;
}

void
Emulator_Manager::increment_emulator_index()
{
    if (p_emulator_list.size() == 0)
        return;

    if (p_current_emulator == p_emulator_list.size() - 1)
        p_current_emulator = 0;
    else
        p_current_emulator++;

    p_listener->emulator_manager_changed();
}

void
Emulator_Manager::decrement_emulator_index()
{
    if (p_emulator_list.size() == 0)
        return;

    if (p_current_emulator == 0)
        p_current_emulator = p_emulator_list.size() - 1;
    else
        p_current_emulator--;

    p_listener->emulator_manager_changed();
}


void
Emulator_Manager::poll_emulator_listener()
{
    p_listener->poll_emulator_manager();
}

Emulator_Manager_Listener::~Emulator_Manager_Listener()
{
}

Emulator_Manager_Listener * Null_Emulator_Manager_Listener::p_instance = 0;

Emulator_Manager_Listener *
Null_Emulator_Manager_Listener::instance()
{
    if (p_instance == 0)
        p_instance = new Null_Emulator_Manager_Listener();

    return p_instance;
}

void
Null_Emulator_Manager_Listener::emulator_manager_changed()
{
}

void
Null_Emulator_Manager_Listener::poll_emulator_manager()
{
}

#define _S(_A_, _B_) \
char * Emulator_Manager:: _A_ = _B_

_S( CONFIG_SECTION,             "Main");
_S( CONFIG_EMULATORS,           "emulators");
_S( CONFIG_PATH_PREFIX,         "path_prefix");

_S( STATE_SECTION,              "Main");
_S( STATE_CURRENT_EMU,          "current_emu");

#undef _S
