#include "view.h"
#include "logfile.h"

using namespace GameLauncher;
using LogFile::gl_log;


Movable_Bitmap::Movable_Bitmap(Engine * engine, int width, int height)
    : p_state(STOPPED)
{
    bool        try_vram;

    p_engine    = engine;

    p_bitmap    = 0;
    p_x         = 0;
    p_y         = 0;
    p_width     = width;
    p_height    = height;
    p_x_velocity        = 0;
    p_y_velocity        = 0;
    p_clip_left         = 0;
    p_clip_top          = 0;
    p_clip_right        = SCREEN_W;
    p_clip_bottom       = SCREEN_H;

    try_vram = ((p_engine->use_vram()) &&
                (p_engine->animation_method() == Engine::PAGE_FLIPPING) &&
                ((gfx_capabilities & GFX_HW_VRAM_BLIT_MASKED) != 0));

    if (try_vram)
    {
        p_bitmap = create_video_bitmap(width, height);
        if (p_bitmap != 0)
            gl_log->info("VRAM bitmap for movable bitmap");
    }

    if (p_bitmap == 0)
    {
        if (try_vram)
            gl_log->info("No VRAM: memory bitmap for moveable bitmap");
        else
            gl_log->info("No HW blit: memory bitmap for movable bitmap");

        p_bitmap = create_bitmap(width, height);
    }
}

Movable_Bitmap::~Movable_Bitmap()
{
    if (p_bitmap != 0)
        destroy_bitmap(p_bitmap);

    return;
}

bool
Movable_Bitmap::update()
{
    if (p_state == STOPPED)
        return false;
    else if (p_state == MOVING)
    {
        bool    moving_right;
        bool    moving_down;

        // Will we reach or pass our final destination?
        moving_right    = (p_x_velocity > 0);
        moving_down     = (p_y_velocity > 0);

        if ((moving_right && ((p_x + p_x_velocity) >= p_move_dest_x)) ||
            ((! moving_right) && ((p_x + p_x_velocity) <= p_move_dest_x)))
        {
            p_x = p_move_dest_x;
        }
        else
            p_x += p_x_velocity;

        if ((moving_down && ((p_y + p_y_velocity) >= p_move_dest_y)) ||
            ((! moving_down) && (p_y + p_y_velocity) <= p_move_dest_y))
        {
            p_y = p_move_dest_y;
        }
        else
            p_y += p_y_velocity;

        if ((p_x == p_move_dest_x) && (p_y == p_move_dest_y))
           p_state = STOPPED;
    }
    
    return true;
}

void
Movable_Bitmap::draw(BITMAP * frame)
{
    fix top;
    fix bottom;
    fix left;
    fix right;
    fix skip_top;
    fix height;
    fix skip_left;
    fix width;

    top         = p_y;
    bottom      = top + p_height - 1;

    if (top < p_clip_top)
        top = p_clip_top;
    if (bottom > p_clip_bottom)
        bottom = p_clip_bottom;

    skip_top = top - p_y;
    height = bottom - top + 1;

    left        = p_x;
    right       = left + p_width - 1;

    if (left < p_clip_left)
        left = p_clip_left;
    if (right > p_clip_right)
        right = p_clip_right;

    skip_left = left - p_x;
    width = right - left + 1;

    masked_blit(p_bitmap, frame, skip_left, skip_top, left, top,
                width, height);

    return;
}

void
Movable_Bitmap::move_to(int x, int y, int milliseconds = 0)
{
    fix       num_game_ticks;

    if (milliseconds == 0)
    {
        p_x = x;
        p_y = y;
        p_state = STOPPED;
        return;
    }

    num_game_ticks = (milliseconds * p_engine->updates_per_second()) / 1000;

    p_move_dest_x = x;
    p_move_dest_y = y;
    p_x_velocity = (p_move_dest_x - p_x) / num_game_ticks;
    p_y_velocity = (p_move_dest_y - p_y) / num_game_ticks;

    // If destination is so close such that the velocity is zero (with
    // rounding), stay stopped
    if ((p_x_velocity == 0) && (p_y_velocity == 0))
        p_state = STOPPED;
    else
        p_state = MOVING;
}

bool
Movable_Bitmap::is_stopped()
{
    return (p_state == STOPPED);
}

void
Movable_Bitmap::set_clip(int left, int top, int right, int bottom)
{
    p_clip_left         = left;
    p_clip_top          = top;
    p_clip_right        = right;
    p_clip_bottom       = bottom;
    return;
}

BITMAP *
Movable_Bitmap::get_bitmap()
{
    return p_bitmap;
}


