#include "view.h"
#include "logfile.h"

#include <stdio.h>

using namespace GameLauncher;
using LogFile::gl_log;

Scrollable_List_VC::Scrollable_List_VC()
    : p_items(), p_visible_items(), p_offscreen_items()
{
}

void
Scrollable_List_VC::initialize(Engine * engine, FONT * font, int x,
                               int y, int width, int height)
{
    p_engine    = engine;
    p_font      = font;
    p_x         = x;
    p_y         = y;
    p_width     = width;
    p_height    = height;
    p_item_height       = text_height(font) + Text_VC::OUTLINE_PADDING;

    p_selected_body_color       = makecol(  0,   0,   0);
    p_selected_outline_color    = makecol(255, 255, 255);
    p_unselected_body_color     = makecol(255, 255, 255);
    p_unselected_outline_color  = makecol(  0,   0,   0);
  
   
    p_use_joystick      =
        p_engine->get_joystick_factory()->is_joystick_detected();
    p_initial_joy_delay = 250;
    p_repeat_joy_delay  = 125;
    p_line_up_joy       = get_repeat_joy(0);
    p_line_down_joy     = get_repeat_joy(0);
    p_page_up_joy       = get_repeat_joy(0);
    p_page_down_joy     = get_repeat_joy(0);
    p_select_joy        = get_one_time_joy(0);
   
    p_mouse_input_axis  = NO_AXIS;
    p_mouse_input_threshold = 1;

    p_selected_index    = 0;
    p_first_shown_index = 0;
    p_initial_key_delay = 250;
    p_repeat_key_delay  = 125; 
    p_line_up_key       = get_repeat_key(0);
    p_line_down_key     = get_repeat_key(0);
    p_page_up_key       = get_repeat_key(0);
    p_page_down_key     = get_repeat_key(0);
    p_select_key        = get_one_time_key(0);
    p_scroll_delay      = SCROLL_DELAY;
    p_turbo_scroll_threshold        = DEFAULT_TURBO_THRESHOLD;
    p_contiguous_page_up_count      = 0;
    p_contiguous_page_down_count    = 0;

    p_null_listener     = new Null_Scrollable_List_Listener;
    p_listener  = p_null_listener;
    p_state     = HIDDEN;

    p_container = new GEContainer();

    create_view_components();
    // Hide 'em all right now
    hide_split(0);

    // Add items to the container
    Text_VC_List::iterator      i;

    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        p_container->add_drawable((*i), DRAW_PRIORITY);
        p_container->add_updatable((*i), UPDATE_PRIORITY);
    }

    for (i = p_offscreen_items.begin(); i != p_offscreen_items.end(); i++)
    {
        p_container->add_drawable((*i), DRAW_PRIORITY);
        p_container->add_updatable((*i), UPDATE_PRIORITY);
    }
}

Scrollable_List_VC::~Scrollable_List_VC()
{
    Text_VC_List::iterator      i;

    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        delete (*i);
    }

    for (i = p_offscreen_items.begin(); i != p_offscreen_items.end(); i++)
    {
        delete (*i);
    }

    delete p_null_listener;
    delete p_container;
}

void
Scrollable_List_VC::draw(BITMAP * frame)
{
    p_container->draw(frame);
}

bool
Scrollable_List_VC::update()
{
    bool        list_updated;
    bool        container_updated;

    list_updated = false;

    container_updated = p_container->update();

    if (p_state == SCROLLING)
    {
        // If the page up/down key is ever physically released, the
        // contiguous count must start over.
        if (! p_page_up_key->is_physically_pressed())
            p_contiguous_page_up_count = 0;

        if (! p_page_down_key->is_physically_pressed())
            p_contiguous_page_down_count = 0;

        if (are_all_items_stopped())
            p_state = UNHIDDEN;
    }

    if (p_state == GOING_INTO_HIDING)
    {
        if (are_all_items_stopped())
            p_state = HIDDEN;
    }

    if (p_state == COMING_OUT_OF_HIDING)
    {
        if (are_all_items_stopped())
            p_state = UNHIDDEN;
    }

    if (p_state == UNHIDDEN)
    {
        list_updated = handle_input();
    }

    return (list_updated || container_updated);
}

bool
Scrollable_List_VC::disable_scroll_delay()
{
    if ((p_turbo_scroll_threshold != -1) &&
        (p_contiguous_page_down_count > p_turbo_scroll_threshold))
    {
        return true;
    }

    if ((p_turbo_scroll_threshold != -1) &&
        (p_contiguous_page_up_count > p_turbo_scroll_threshold))
    {
        return true;
    }

    if ((p_last_input_command == MOUSE_MOVE_UP) ||
        (p_last_input_command == MOUSE_MOVE_DOWN))
    {
        return true;
    }

    return false;
}

bool
Scrollable_List_VC::handle_input()
{
    bool    updated;
    int     mouse_quantum;
    unsigned int    move_up;
    unsigned int    move_down;
    unsigned int    scroll_up;
    unsigned int    scroll_down;
    unsigned int    last_shown_index;
    Input_Command   this_command;


    if (p_items.size() == 0)
        return false;

    if (p_select_key->is_pressed() || p_select_joy->is_pressed())
    {
        p_listener->item_selected(p_selected_index);
        return false;
    }

    updated         = false;
    move_up         = 0;
    move_down       = 0;
    scroll_up       = 0;
    scroll_down     = 0;
    this_command    = NOTHING;

    // Calculate last_shown_index, clipping if necessary
    last_shown_index = p_first_shown_index + p_visible_items.size() - 1;
    if (last_shown_index >= p_items.size())
        last_shown_index = p_items.size() - 1;

    if (p_line_down_key->is_pressed() || p_line_down_joy->is_pressed())
    {
        move_down = 1;
        if (p_selected_index == last_shown_index)
            scroll_down = 1;
        this_command = LINE_DOWN;
    }

    if (p_line_up_key->is_pressed() || p_line_up_joy->is_pressed())
    {
        move_up = 1;
        if (p_selected_index == p_first_shown_index)
            scroll_up = 1;
        this_command = LINE_UP;
    }

    if (p_page_up_key->is_pressed() || p_page_up_joy->is_pressed())
    {
        if (p_selected_index == p_first_shown_index)
        {
            // At top, scroll up a page.
            move_up = p_visible_items.size() - 1;
            scroll_up = p_visible_items.size() - 1;
        }
        else
        {
            // Move to first shown.
            move_up = p_selected_index - p_first_shown_index;
        }

        this_command = PAGE_UP;
    }


    if (p_page_down_key->is_pressed() || p_page_down_joy->is_pressed())
    {
        if (p_selected_index == last_shown_index)
        {
            // At end, scroll down a page.
            move_down = p_visible_items.size() - 1;
            scroll_down = p_visible_items.size() - 1;
        }
        else
        {
            // Move to last shown
            move_down = last_shown_index - p_selected_index;
        }

        this_command = PAGE_DOWN;
    }

    mouse_quantum = get_mouse_input_quantum();
    if (mouse_quantum < 0)
    {
        move_up = 1;
        if (p_selected_index == p_first_shown_index)
            scroll_up = 1;
        this_command = MOUSE_MOVE_UP;
    }
    else if (mouse_quantum > 0)
    {
        move_down = 1;
        if (p_selected_index == last_shown_index)
            scroll_down = 1;
        this_command = MOUSE_MOVE_DOWN;
    }

    if (this_command != NOTHING)
    {
        // A contiguous page up/down occurs if we scrolled last time,
        // we're going to scroll again in the same direction this
        // time, and it was an artificial keypress.  If it's a real
        // key press, but we're scrolling, it counts as "1" contiguous
        // scroll.

        if ((p_last_input_command == PAGE_UP) && (this_command == PAGE_UP) &&
            (scroll_up > 1))
        {
            if (p_page_up_key->is_artificial())
                p_contiguous_page_up_count++;
            else
                p_contiguous_page_up_count = 1;
        }
        else
        {
            p_contiguous_page_up_count = 0;
        }

        if ((p_last_input_command == PAGE_DOWN) &&
            (this_command == PAGE_DOWN) &&
            (scroll_down > 1))
        {
            if (p_page_down_key->is_artificial())
                p_contiguous_page_down_count++;
            else
                p_contiguous_page_down_count = 1;
        }
        else
        {
            p_contiguous_page_down_count = 0;
        }

        p_last_input_command = this_command;
    }
    else
    {

        // If the page up/down key is ever physically released, the
        // contiguous count must start over.
        if (! p_page_up_key->is_physically_pressed())
            p_contiguous_page_up_count = 0;

        if (! p_page_down_key->is_physically_pressed())
            p_contiguous_page_down_count = 0;
    }

    // Any wierd combination where up and down movement is
    // involved results in no effect
    if ((move_up > 0) && (move_down > 0))
    {
        move_down       = 0;
        move_up         = 0;
    }

    // Make sure we don't move up past the beginning.
    if ((move_up > 0) && (p_selected_index < move_up))
    {
        move_up     = p_selected_index;
        scroll_up   = p_selected_index;
    }

    // Make sure we don't move down past the end.
    if ((move_down > 0) &&
        ((p_selected_index + move_down) > (p_items.size() - 1)))
    {
        move_down   = p_items.size() - 1 - p_selected_index;
        scroll_down = p_items.size() - 1 - p_selected_index;
    }


    if ((move_up > 0) || (move_down > 0))
    {
        int    saved_scroll_delay;
        
        saved_scroll_delay  = p_scroll_delay;
        
        if (disable_scroll_delay())
        {
            p_scroll_delay = 0;
        }

        (*p_selected_item)->set_colors(p_unselected_body_color,
                                       p_unselected_outline_color);

        if (move_up > 0)
        {
            move_selection_up(move_up, scroll_up);
        }
        else if (move_down > 0)
        {
            move_selection_down(move_down, scroll_down);
        }

        (*p_selected_item)->set_colors(p_selected_body_color,
                                       p_selected_outline_color);

        p_listener->selection_changed(p_selected_index);

        updated = true;

        p_scroll_delay = saved_scroll_delay;
    }

    return updated;
}

int
Scrollable_List_VC::get_mouse_input_quantum()
{
    int quantum;
    int remainder;
    int current_mouse;
    int direction;

    switch(p_mouse_input_axis)
    {
        case POSITIVE_X_AXIS:
            current_mouse   = p_engine->get_infinite_mouse_x();
            direction       = 1;
            break;

        case NEGATIVE_X_AXIS:
            current_mouse   = p_engine->get_infinite_mouse_x();
            direction       = -1;
            break;

        case POSITIVE_Y_AXIS:
            current_mouse   = p_engine->get_infinite_mouse_y();
            direction       = 1;
            break;

        case NEGATIVE_Y_AXIS:
            current_mouse   = p_engine->get_infinite_mouse_y();
            direction       = -1;
            break;

        default:
            current_mouse   = 0;
            direction       = 1;
    }

    // Keep infinite value from ever rolling over by constantly
    // mod'ing it against the threshold.  Also, set the other infinte
    // axis to 0 to keep it from rolling over.

    quantum     = current_mouse / p_mouse_input_threshold;
    remainder   = current_mouse % p_mouse_input_threshold;

    switch(p_mouse_input_axis)
    {
        case POSITIVE_X_AXIS:
        case NEGATIVE_X_AXIS:
            p_engine->set_infinite_mouse_x(remainder);
            p_engine->set_infinite_mouse_y(0);
            break;

        case POSITIVE_Y_AXIS:
        case NEGATIVE_Y_AXIS:
            p_engine->set_infinite_mouse_x(0);
            p_engine->set_infinite_mouse_y(remainder);
            break;

        default:
            // Do nothing
            break;
    }

    return (quantum * direction);
}

void
Scrollable_List_VC::move_selection_up(unsigned move, unsigned scroll)
{
    unsigned    i;
    int         scroll_ms;
    int         screen_index;
    int         x;
    int         y;
    Text_VC_List        moved_items;

    scroll_ms = p_scroll_delay * ((scroll + 1)/ 2);

    // First, perform the scroll
    for (i = 0; i < scroll; i++)
    {
        Text_VC *       item;
        unsigned        next_index;

        // Remove last item, and stick it offscreen.
        item = p_visible_items.back();
        p_visible_items.pop_back();

        p_offscreen_items.push_front(item);
        moved_items.push_front(item);

        // Pull a selection from the offscreen list, setup the new text,
        // and move it to the top of the screen.
        item = p_offscreen_items.back();
        p_offscreen_items.pop_back();

        next_index      = p_selected_index - i - 1;
        screen_index    = 0 - int(i) - 1;
        get_index_coordinates(screen_index, &x, &y);

        // Allow listener to change contents for lazy setting of items.
        p_listener->item_visible(next_index);

        item->set_contents(p_items[next_index].text, p_items[next_index].icon);
        item->move_to(x, y);
        // item->move_to(p_x, p_y - ((i+1)*p_item_height));
        p_visible_items.push_front(item);

        p_first_shown_index--;
        p_engine->poll_background_task();
    }

    // Next, perform move.  Must be done in a loop becuase
    // p_selections is not random access.
    for (i = 0; i < move; i++)
    {
        p_selected_index--;
        p_selected_item--;
    }

    // Move positions
    if (scroll > 0)
    {
        Text_VC_List::iterator  j;

        screen_index = 0;
        for (j = p_visible_items.begin(); j != p_visible_items.end(); j++)
        {
            get_index_coordinates(screen_index, &x, &y);
            (*j)->move_to(x, y, scroll_ms);
            screen_index++;
        }

        // Put moved visible items past the bottom of the list
        for (j = moved_items.begin(); j != moved_items.end(); j++)
        {
            get_index_coordinates(screen_index, &x, &y);
            (*j)->move_to(x, y, scroll_ms);
            screen_index++;
        }
        p_state = SCROLLING;
    }

    return;
}

void
Scrollable_List_VC::move_selection_down(unsigned move, unsigned scroll)
{
    unsigned    i;
    int         scroll_ms;
    int         screen_index;
    int         x;
    int         y;
    Text_VC_List        moved_items;

    scroll_ms = p_scroll_delay * ((scroll + 1)/ 2);

    // First, perform the scroll
    for (i = 0; i < scroll; i++)
    {
        Text_VC *       item;
        unsigned        next_index;

        // Remove first item, but stick it on the back of the offscreen
        // list.
        item    = p_visible_items.front();
        p_visible_items.pop_front();
        p_offscreen_items.push_back(item);
        moved_items.push_back(item);

        // Pull a selection off the front of the offscreen list, setup
        // the new text, and move it to the bottom.
        item    = p_offscreen_items.front();
        p_offscreen_items.pop_front();

        next_index      = p_selected_index + i + 1;
        screen_index    = p_num_visible_items + i;
        get_index_coordinates(screen_index, &x, &y);

        // Allow listener to change contents for lazy setting of items.
        p_listener->item_visible(next_index);

        item->set_contents(p_items[next_index].text, p_items[next_index].icon);
        item->move_to(x, y);
        p_visible_items.push_back(item);

        p_first_shown_index++;
        p_engine->poll_background_task();
    }

    // Next, perform move.  Must be done in a loop becuase
    // p_selections is not random access.
    for (i = 0; i < move; i++)
    {
        p_selected_index++;
        p_selected_item++;
    }

    // Move positions
    if (scroll > 0)
    {
        Text_VC_List::reverse_iterator  j;

        screen_index = p_num_visible_items - 1;
        for (j = p_visible_items.rbegin(); j != p_visible_items.rend(); j++)
        {
            get_index_coordinates(screen_index, &x, &y);
            (*j)->move_to(x, y, scroll_ms);
            screen_index--;
        }

        for (j = moved_items.rbegin(); j != moved_items.rend(); j++)
        {
            get_index_coordinates(screen_index, &x, &y);
            (*j)->move_to(x, y, scroll_ms);
            screen_index--;
        }

        p_state = SCROLLING;
    }

    return;
}

void
Scrollable_List_VC::update_visible_items()
{
    Text_VC_List::iterator      i;
    unsigned    index;
    unsigned    num_items;

    num_items   = p_items.size();
    index       = p_first_shown_index;

    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        if (index >= num_items)
        {
            (*i)->set_contents("", 0);
            (*i)->set_colors(p_unselected_body_color,
                             p_unselected_outline_color);
            continue;
        }

        // Give listener a chance to update the item.
        p_listener->item_visible(index);

        (*i)->set_contents(p_items[index].text, p_items[index].icon);
        if (index == p_selected_index)
        {
            (*i)->set_colors(p_selected_body_color, p_selected_outline_color);
            p_selected_item = i;
        }
        else
        {
            (*i)->set_colors(p_unselected_body_color,
                             p_unselected_outline_color);
        }
        index++;
    }
}

void
Scrollable_List_VC::update_offscreen_items()
{
    Text_VC_List::iterator      i;

    for (i = p_offscreen_items.begin(); i != p_offscreen_items.end(); i++)
    {
        (*i)->set_colors(p_unselected_body_color,
                         p_unselected_outline_color);
    }
}

void
Scrollable_List_VC::set_selected_colors(int body, int outline)
{
    p_selected_body_color       = body;
    p_selected_outline_color    = outline;
    update_visible_items();
    update_offscreen_items();
}

void
Scrollable_List_VC::set_unselected_colors(int body, int outline)
{
    p_unselected_body_color     = body;
    p_unselected_outline_color  = outline;
    update_visible_items();
    update_offscreen_items();
}

void
Scrollable_List_VC::set_all_items(const Scrollable_Item_List & items,
                                  unsigned selected_index,
                                  unsigned first_shown_index)
{
    unsigned    last_items_index;

    p_items             = items;

    // It's ok to have the last index be 0 if the list is emtpy.  All
    // other areas of the program need to handle this.
    if (p_items.size() == 0)
        last_items_index    = 0;
    else
        last_items_index    = p_items.size() - 1;

    p_first_shown_index = first_shown_index;
    if (p_first_shown_index > last_items_index)
        p_first_shown_index = last_items_index;

    p_selected_index    = selected_index;
    if (p_selected_index > last_items_index)
        p_selected_index = last_items_index;

    // Make sure the selected index is one of the visible items, and
    // correct if not.
    if (p_selected_index < p_first_shown_index)
        p_selected_index = p_first_shown_index;

    if (p_selected_index > (p_first_shown_index + p_num_visible_items - 1))
        p_selected_index = p_first_shown_index + p_num_visible_items - 1;

    update_visible_items();
}

void
Scrollable_List_VC::set_item_at(const Scrollable_Item & item,
                                unsigned index)
{
    if (index < p_items.size())
        p_items[index] = item;
}

unsigned
Scrollable_List_VC::get_selected_index()
{
    return p_selected_index;
}

unsigned
Scrollable_List_VC::get_first_shown_index()
{
    return p_first_shown_index;
}

const Scrollable_Item *
Scrollable_List_VC::get_item_at(unsigned index)
{
    return &p_items[index];
}

void
Scrollable_List_VC::set_listener(Scrollable_List_Listener * listener)
{
    if (listener != 0)
        p_listener      = listener;
    else
        p_listener      = p_null_listener;
}

Joystick_Watcher *
Scrollable_List_VC::get_one_time_joy(int joystick_event)
{
    Joystick_Watcher *  watcher;
    Joystick_Event *    event;

    event   = get_joystick_event(joystick_event);
    watcher = p_engine->get_joystick_factory()->create_joystick_watcher(event);
    watcher->set_one_time();
    
    return watcher;
}

Joystick_Watcher *
Scrollable_List_VC::get_repeat_joy(int joystick_event)
{
    Joystick_Watcher *  watcher;
    Joystick_Event *    event;

    event   = get_joystick_event(joystick_event);
    watcher = p_engine->get_joystick_factory()->create_joystick_watcher(event);
    watcher->set_repeat(p_initial_joy_delay, p_repeat_joy_delay);
    
    return watcher;
}

void
Scrollable_List_VC::set_joystick_repeat_delay(int initial, int repeat)
{
    p_initial_joy_delay = initial;
    p_repeat_joy_delay  = repeat;

    p_line_up_joy->set_repeat(p_initial_joy_delay, p_repeat_joy_delay);
    p_line_down_joy->set_repeat(p_initial_joy_delay, p_repeat_joy_delay);
    p_page_up_joy->set_repeat(p_initial_joy_delay, p_repeat_joy_delay);
    p_page_down_joy->set_repeat(p_initial_joy_delay, p_repeat_joy_delay);
}

Joystick_Event *
Scrollable_List_VC::get_joystick_event(int joystick_event)
{
    Joystick_Event *    e;
    Joystick_Digital_Axis_Event::Position   p1;
    Joystick_Digital_Axis_Event::Position   p2;

    if (! p_use_joystick)
        return 0;

    p1  = Joystick_Digital_Axis_Event::POSITION_ONE;
    p2  = Joystick_Digital_Axis_Event::POSITION_TWO;

    switch(joystick_event)
    {
        case BUTTON_A:
            e = new Joystick_Button_Event(0, 0);
            break;
            
        case BUTTON_B:
            e = new Joystick_Button_Event(0, 1);
            break;
            
        case JOY_UP:
            e = new Joystick_Digital_Axis_Event(0, 0, 1, p1);
            break;
            
        case JOY_DOWN:
            e = new Joystick_Digital_Axis_Event(0, 0, 1, p2);
            break;
            
        case JOY_LEFT:
            e = new Joystick_Digital_Axis_Event(0, 0, 0, p1);
            break;
            
        case JOY_RIGHT:
            e = new Joystick_Digital_Axis_Event(0, 0, 0, p2);
            break;
            
        default:
            e = 0;
    }

    return e;
}

void
Scrollable_List_VC::set_select_joy(int joystick_event)
{
    p_select_joy->set_joystick_event(get_joystick_event(joystick_event));
}

void
Scrollable_List_VC::set_line_down_joy(int joystick_event)
{
    p_line_down_joy->set_joystick_event(get_joystick_event(joystick_event));
}

void
Scrollable_List_VC::set_line_up_joy(int joystick_event)
{
    p_line_up_joy->set_joystick_event(get_joystick_event(joystick_event));
}

void
Scrollable_List_VC::set_page_up_joy(int joystick_event)
{
    p_page_up_joy->set_joystick_event(get_joystick_event(joystick_event));
}

void
Scrollable_List_VC::set_page_down_joy(int joystick_event)
{
    p_page_down_joy->set_joystick_event(get_joystick_event(joystick_event));
}

Keypress_Watcher *
Scrollable_List_VC::get_repeat_key(int scancode)
{
    Keypress_Watcher * watcher;
    watcher = 
        p_engine->get_keypress_factory()->create_keypress_watcher(scancode);
    watcher->set_repeat(p_initial_key_delay, p_repeat_key_delay);
    return watcher;
}

Keypress_Watcher *
Scrollable_List_VC::get_one_time_key(int scancode)
{
    Keypress_Watcher * watcher;
    watcher = 
        p_engine->get_keypress_factory()->create_keypress_watcher(scancode);
    watcher->set_one_time();
    return watcher;
}

void
Scrollable_List_VC::set_line_up_key(int scancode)
{
    p_line_up_key->set_scancode(scancode);
}

void
Scrollable_List_VC::set_line_down_key(int scancode)
{
    p_line_down_key->set_scancode(scancode);
}

void
Scrollable_List_VC::set_page_up_key(int scancode)
{
    p_page_up_key->set_scancode(scancode);
}

void
Scrollable_List_VC::set_page_down_key(int scancode)
{
    p_page_down_key->set_scancode(scancode);
}

void
Scrollable_List_VC::set_select_key(int scancode)
{
    p_select_key->set_scancode(scancode);
}

void
Scrollable_List_VC::set_key_repeat_delay(int initial, int repeat)
{
    p_initial_key_delay = initial;
    p_repeat_key_delay  = repeat;

    p_line_up_key->set_repeat(p_initial_key_delay, p_repeat_key_delay);
    p_line_down_key->set_repeat(p_initial_key_delay, p_repeat_key_delay);
    p_page_up_key->set_repeat(p_initial_key_delay, p_repeat_key_delay);
    p_page_down_key->set_repeat(p_initial_key_delay, p_repeat_key_delay);
}

void
Scrollable_List_VC::set_mouse_input_axis(Mouse_Input_Axis axis)
{
    p_mouse_input_axis = axis;
}

void
Scrollable_List_VC::set_mouse_input_threshold(int threshold)
{
    p_mouse_input_threshold = threshold;
}

void
Scrollable_List_VC::set_use_animations(bool use_animations)
{
    if (use_animations)
        p_scroll_delay  = SCROLL_DELAY;
    else
        p_scroll_delay  = 0;
}

void
Scrollable_List_VC::set_turbo_scroll_threshold(int threshold)
{
    p_turbo_scroll_threshold    = threshold;
}

bool
Scrollable_List_VC::are_all_items_stopped()
{
    Text_VC_List::iterator      i;
    bool        all_stopped;

    all_stopped = true;
    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        if (! (*i)->is_stopped())
        {
            all_stopped = false;
            break;
        }
    }

    return all_stopped;
}


void
Scrollable_List_VC::get_index_coordinates(int index, int * x, int * y)
{
    *x  = p_visible_top_x + index * p_visible_delta_x;
    *y  = p_visible_top_y + index * p_visible_delta_y;
}


void
Scrollable_List_VC::hide_offscreen_items()
{
    Text_VC_List::iterator      i;

    for (i = p_offscreen_items.begin(); i != p_offscreen_items.end(); i++)
    {
        (*i)->move_to(p_offscreen_hidden_x, p_offscreen_hidden_y);
    }
}

void
Scrollable_List_VC::hide_split(int milliseconds)
{
    Text_VC_List::iterator      i;
    int current_row;
    int hidden_left_x;
    int hidden_left_y;
    int hidden_right_x;
    int hidden_right_y;

    hidden_left_x       = p_visible_hidden_left_x;
    hidden_left_y       = p_visible_hidden_left_y;
    hidden_right_x      = p_visible_hidden_right_x;
    hidden_right_y      = p_visible_hidden_right_y;

    current_row = 0;
    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        if ((current_row % 2) == 0)
            (*i)->move_to(hidden_left_x, hidden_left_y, milliseconds);
        else
            (*i)->move_to(hidden_right_x, hidden_right_y, milliseconds);
            
        hidden_left_x   += p_visible_delta_x;
        hidden_left_y   += p_visible_delta_y;
        hidden_right_x  += p_visible_delta_x;
        hidden_right_y  += p_visible_delta_y;
        current_row++;
    }

    hide_offscreen_items();
    p_state = GOING_INTO_HIDING;
}

void
Scrollable_List_VC::hide_left(int milliseconds)
{
    Text_VC_List::iterator      i;
    int hidden_left_x;
    int hidden_left_y;

    hidden_left_x       = p_visible_hidden_left_x;
    hidden_left_y       = p_visible_hidden_left_y;
    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        (*i)->move_to(hidden_left_x, hidden_left_y, milliseconds);
        hidden_left_x   += p_visible_delta_x;
        hidden_left_y   += p_visible_delta_y;
    }

    hide_offscreen_items();
    p_state = GOING_INTO_HIDING;
}

void
Scrollable_List_VC::hide_right(int milliseconds)
{
    Text_VC_List::iterator      i;
    int hidden_right_x;
    int hidden_right_y;

    hidden_right_x      = p_visible_hidden_right_x;
    hidden_right_y      = p_visible_hidden_right_y;
    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        (*i)->move_to(hidden_right_x, hidden_right_y, milliseconds);
        hidden_right_x  += p_visible_delta_x;
        hidden_right_y  += p_visible_delta_y;
    }

    hide_offscreen_items();
    p_state = GOING_INTO_HIDING;
}

void
Scrollable_List_VC::center_selected_hide_others(int milliseconds)
{
    Text_VC_List::iterator      i;
    unsigned    current_index;
    int current_row;
    int hidden_left_x;
    int hidden_left_y;
    int hidden_right_x;
    int hidden_right_y;

    hidden_left_x       = p_visible_hidden_left_x;
    hidden_left_y       = p_visible_hidden_left_y;
    hidden_right_x      = p_visible_hidden_right_x;
    hidden_right_y      = p_visible_hidden_right_y;

    current_row         = 0;
    current_index       = p_first_shown_index;

    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        if (current_index == p_selected_index)
            (*i)->move_to(p_centered_x, p_centered_y, milliseconds);
        else if ((current_row % 2) == 0)
            (*i)->move_to(hidden_left_x, hidden_left_y, milliseconds);
        else
            (*i)->move_to(hidden_right_x, hidden_right_y, milliseconds);
            
        hidden_left_x   += p_visible_delta_x;
        hidden_left_y   += p_visible_delta_y;
        hidden_right_x  += p_visible_delta_x;
        hidden_right_y  += p_visible_delta_y;
        current_row++;
        current_index++;
    }

    hide_offscreen_items();
    p_state = GOING_INTO_HIDING;
}

void
Scrollable_List_VC::unhide_all(int milliseconds)
{
    Text_VC_List::iterator      i;
    int visible_x;
    int visible_y;

    visible_x   = p_visible_top_x;
    visible_y   = p_visible_top_y;
    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        (*i)->move_to(visible_x, visible_y, milliseconds);
        visible_x       += p_visible_delta_x;
        visible_y       += p_visible_delta_y;
    }
    p_state = COMING_OUT_OF_HIDING;
}



void
Null_Scrollable_List_Listener::selection_changed(unsigned index)
{
}

void
Null_Scrollable_List_Listener::item_selected(unsigned index)
{
}

void
Null_Scrollable_List_Listener::item_visible(unsigned index)
{
}


/************************************************************************
 *
 * Scrollable_HList_VC
 *
 ************************************************************************/

Scrollable_HList_VC::Scrollable_HList_VC(Engine * engine, FONT * font, int x,
                                         int y, int width, int height)
    : Scrollable_List_VC()
{
    initialize(engine, font, x, y, width, height);
}

void
Scrollable_HList_VC::create_view_components()
{
    int visible_bottom_y;
    int i;

    p_visible_top_x     = p_x;
    p_visible_top_y     = p_y;
    p_visible_delta_x   = 0;
    p_visible_delta_y   = p_item_height;

    p_visible_hidden_left_x     = 0 - SCREEN_W + p_x;
    p_visible_hidden_left_y     = p_y;

    p_visible_hidden_right_x    = SCREEN_W + p_x;
    p_visible_hidden_right_y    = p_y;

    p_offscreen_hidden_x        = p_x;
    p_offscreen_hidden_y        = 0 - p_item_height;

    p_num_visible_items = p_height / p_item_height;
    visible_bottom_y    = p_y + p_num_visible_items * p_item_height;

    p_centered_x        = p_x;
    p_centered_y        = p_y + p_height/2;

    // Create the visible text items
    for (i = 0; i < p_num_visible_items; i++)
    {
        HText_VC *      text_component;

        text_component  = new HText_VC(p_engine, p_width, p_item_height,
                                       p_font);
        text_component->set_clip(0, p_visible_top_y,
                                 SCREEN_W, visible_bottom_y);
        p_visible_items.push_back(text_component);
    }

    // Create the off screen text items
    for (i = 0; i < p_num_visible_items; i++)
    {
        HText_VC *      text_component;

        text_component  = new HText_VC(p_engine, p_width, p_item_height,
                                       p_font);
        text_component->set_clip(0, p_y,
                                 SCREEN_W, visible_bottom_y);
        p_offscreen_items.push_back(text_component);
    }
}


/************************************************************************
 *
 * Scrollable_VLList_VC
 *
 ************************************************************************/

Scrollable_VLList_VC::Scrollable_VLList_VC(Engine * engine, FONT * font, int x,
                                           int y, int width, int height)
    : Scrollable_List_VC()
{
    initialize(engine, font, x, y, width, height);
}

void
Scrollable_VLList_VC::create_view_components()
{
    int visible_bottom_x;
    int i;

    p_visible_top_x     = p_x;
    p_visible_top_y     = p_y;
    p_visible_delta_x   = p_item_height;
    p_visible_delta_y   = 0;

    p_visible_hidden_left_x     = p_x;
    p_visible_hidden_left_y     = SCREEN_H + p_y;

    p_visible_hidden_right_x    = p_x;
    p_visible_hidden_right_y    = 0 - SCREEN_H + p_y;

    p_offscreen_hidden_x        = 0 - p_item_height;
    p_offscreen_hidden_y        = p_y;

    p_num_visible_items = p_width / p_item_height;
    visible_bottom_x    = p_x + p_num_visible_items * p_item_height;

    p_centered_x        = p_x + p_width/2;
    p_centered_y        = p_y;

    // Create the visible text items
    for (i = 0; i < p_num_visible_items; i++)
    {
        VText_VC *      text_component;
        
        text_component  = new VText_VC(p_engine, p_item_height, p_height,
                                       p_font);
        text_component->set_clip(p_x, 0,
                                 visible_bottom_x, SCREEN_H);
        p_visible_items.push_back(text_component);
    }

    // Create the off screen text items
    for (i = 0; i < p_num_visible_items; i++)
    {
        VText_VC *      text_component;

        text_component  = new VText_VC(p_engine, p_item_height, p_height,
                                       p_font);
        text_component->set_clip(p_x, 0,
                                 visible_bottom_x, SCREEN_H);
        p_offscreen_items.push_back(text_component);
    }
}
